"""
========================================================================
Bits.py
========================================================================
Pure-Python implementation of fixed-bitwidth data type.

Author : Shunning Jiang
Date   : Oct 31, 2017
"""

# lower <= value <= upper
_upper = [ 0,  1 ]
_lower = [ 0, -1 ]

for i in range(2, 1024):
  _upper.append( (_upper[i-1] << 1) + 1 )
  _lower.append(  _lower[i-1] << 1      )

object_new = object.__new__
def _new_valid_bits( nbits, uint ):
  ret = object_new( Bits )
  ret._nbits = nbits
  ret._uint  = uint
  return ret

class Bits:
  __slots__ = ( "_nbits", "_uint", "_next" )

  @property
  def nbits( self ):
    return self._nbits

  def __init__( self, nbits, v=0, trunc_int=False ):
    nbits = int(nbits)
    if nbits < 1 or nbits >= 1024: raise ValueError(f"Only support 1 <= nbits < 1024, not {nbits}")

    self._nbits = nbits

    if isinstance( v, Bits ):
      if nbits != v.nbits:
        if nbits < v.nbits:
          raise ValueError( f"The Bits{v.nbits} object on RHS is too wide to be used to construct Bits{nbits}!\n"
                            f"- Suggestion: directly use trunc( value, {nbits}/Bits{nbits} )" )
        else:
          raise ValueError( f"The Bits{v.nbits} object on RHS is too narrow to be used to construct Bits{nbits}!\n"
                            f"- Suggestion: directly use zext/sext(value, {nbits}/Bits{nbits} )" )
      self._uint = v._uint
    else:
      v = int(v)
      up = _upper[nbits]

      if not trunc_int:
        lo = _lower[nbits]
        if v < lo or v > up:
          raise ValueError( f"Value {hex(v)} is too wide for Bits{nbits}!\n" \
                            f"(Bits{nbits} only accepts {hex(lo)} <= value <= {hex(up)})" )
      self._uint = v & up

  # PyMTL simulation specific

  def __ilshift__( self, v ):
    nbits = self._nbits
    try:
      # Bits/Bitstruct
      if v.nbits != nbits:
        if v.nbits < nbits:
          raise ValueError( f"Bitwidth of LHS must be equal to RHS during <<= non-blocking assignment, " \
                            f"but here LHS Bits{nbits} > RHS Bits{v.nbits}.\n"
                            f"- Suggestion: LHS @= zext/sext(RHS, nbits/Type)" )
        else:
          raise ValueError( f"Bitwidth of LHS must be equal to RHS during <<= non-blocking assignment, " \
                            f"but here LHS Bits{nbits} < RHS Bits{v.nbits}.\n"
                            f"- Suggestion: LHS @= trunc(RHS, nbits/Type)" )
      self._next = v.to_bits()._uint
    except AttributeError:
      # Cast to int
      v = int(v)
      lo = _lower[nbits]
      up = _upper[nbits]

      if v < lo or v > up:
        raise ValueError( f"RHS value {hex(v)} of <<= is too wide for LHS Bits{nbits}!\n" \
                          f"(Bits{nbits} only accepts {hex(lo)} <= value <= {hex(up)})" )
      self._next = v & up

    return self

  def _flip( self ):
    self._uint = self._next

  def clone( self ):
    return _new_valid_bits( self._nbits, self._uint )

  def __deepcopy__( self, memo ):
    return _new_valid_bits( self._nbits, self._uint )

  def __imatmul__( self, v ):
    nbits = self._nbits
    try:
      # Bits/Bitstruct
      if v.nbits != nbits:
        if v.nbits < nbits:
          raise ValueError( f"Bitwidth of LHS must be equal to RHS during @= blocking assignment, " \
                            f"but here LHS Bits{nbits} > RHS Bits{v.nbits}.\n"
                            f"- Suggestion: LHS @= zext/sext(RHS, nbits/Type)" )
        else:
          raise ValueError( f"Bitwidth of LHS must be equal to RHS during @= blocking assignment, " \
                            f"but here LHS Bits{nbits} < RHS Bits{v.nbits}.\n"
                            f"- Suggestion: LHS @= trunc(RHS, nbits/Type)" )
      self._uint = v.to_bits()._uint
    except AttributeError:
      # Cast to int
      v = int(v)

      lo = _lower[nbits]
      up = _upper[nbits]

      if v < lo or v > up:
        raise ValueError( f"RHS value {hex(v)} of @= is too wide for LHS Bits{nbits}!\n" \
                          f"(Bits{nbits} only accepts {hex(lo)} <= value <= {hex(up)})" )
      self._uint = v & up

    return self

  def to_bits( self ):
    return self

  # Arithmetics
  def __getitem__( self, idx ):

    if isinstance( idx, slice ):
      if idx.step:
        raise IndexError( "Index cannot contain step" )
      try:
        start, stop = int(idx.start or 0), int(idx.stop or self._nbits)
        assert 0 <= start < stop <= self._nbits
      except:
        raise IndexError( f"Invalid access: [{idx.start}:{idx.stop}] in a Bits{self._nbits} instance" )

      # Bypass check
      nbits = stop - start
      return _new_valid_bits( stop-start, (self._uint >> start) & _upper[nbits] )

    i = int(idx)
    if i >= self._nbits or i < 0:
      raise IndexError( f"Invalid access: [{i}] in a Bits{self._nbits} instance" )

    # Bypass check
    return _new_valid_bits( 1, (self._uint >> i) & 1 )

  def __setitem__( self, idx, v ):
    sv = int(self._uint)

    if isinstance( idx, slice ):
      if idx.step:
        raise IndexError( "Index cannot contain step" )
      try:
        start, stop = int(idx.start or 0), int(idx.stop or self._nbits)
        assert 0 <= start < stop <= self._nbits
      except:
        raise IndexError( f"Invalid access: [{idx.start}:{idx.stop}] in a Bits{self._nbits} instance" )

      slice_nbits = stop - start
      if isinstance( v, Bits ):
        if v.nbits != slice_nbits:
          if v.nbits < slice_nbits:
            raise ValueError( f"Cannot fit a Bits{v.nbits} object into a {slice_nbits}-bit slice [{start}:{stop}]\n"
                              f"- Suggestion: sext/zext the RHS")
          else:
            raise ValueError( f"Cannot fit a Bits{v.nbits} object into a {slice_nbits}-bit slice [{start}:{stop}]\n"
                              f"- Suggestion: trunc the RHS")

        self._uint = (sv & (~((1 << stop) - (1 << start)))) | \
                     ((v._uint & _upper[slice_nbits]) << start)
      else:
        # Cast to int
        v = int(v)
        lo = _lower[slice_nbits]
        up = _upper[slice_nbits]

        if v < lo or v > up:
          raise ValueError( f"Cannot fit {v} into a Bits{slice_nbits} slice\n" \
                            f"(Bits{slice_nbits} only accepts {hex(lo)} <= value <= {hex(up)})" )

        self._uint = (sv & (~((1 << stop) - (1 << start)))) | \
                     ((v & _upper[slice_nbits]) << start)
      return

    i = int(idx)
    if i >= self._nbits or i < 0:
      raise IndexError( f"Invalid access: [{i}] in a Bits{self._nbits} instance" )

    if isinstance( v, Bits ):
      if v.nbits > 1:
        raise ValueError( f"Cannot fit a Bits{v.nbits} object into the 1-bit slice" )
      self._uint = (sv & ~(1 << i)) | ((v._uint & 1) << i)
    else:
      v = int(v)
      if abs(v) > 1:
        raise ValueError( f"Value {hex(v)} is too big for the 1-bit slice!\n" )
      self._uint = (sv & ~(1 << i)) | ((int(v) & 1) << i)

  def __add__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '+' (add) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( nbits, (self._uint + other._uint) & _upper[nbits] )
    except AttributeError:
      other = int(other)
      up = _upper[ nbits ]
      if other < 0 or other > up:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(up)}" )
      return _new_valid_bits( nbits, (self._uint + other) & up )

  def __radd__( self, other ):
    return self.__add__( other )

  def __sub__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '-' (sub) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( nbits, (self._uint - other._uint) & _upper[nbits] )
    except AttributeError:
      other = int(other)
      up = _upper[ nbits ]
      if other < 0 or other > up:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(up)}" )
      return _new_valid_bits( nbits, (self._uint - other) & up )

  def __rsub__( self, other ):
    nbits = self._nbits
    # Shouldn't be Bits
    other = int(other)
    up = _upper[ nbits ]
    if other < 0 or other > up:
      raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                        f"Suggestion: 0 <= x <= {hex(up)}" )
    return _new_valid_bits( nbits, (other - self._uint) & up )

  def __mul__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '*' (mul) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( nbits, (self._uint * other._uint) & _upper[nbits] )
    except AttributeError:
      other = int(other)
      up = _upper[ nbits ]
      if other < 0 or other > up:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(up)}" )
      return _new_valid_bits( nbits, (self._uint * other) & up)

  def __rmul__( self, other ):
    return self.__mul__( other )

  # no need to AND mask
  def __and__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '&' (and) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( nbits, self._uint & other._uint )
    except AttributeError:
      other = int(other)
      if other < 0 or other > _upper[ nbits ]:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(_upper[ nbits ])}" )
      return _new_valid_bits( nbits, self._uint & other )

  def __rand__( self, other ):
    return self.__and__( other )

  # no need to AND mask
  def __or__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '|' (or) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( nbits, self._uint | other._uint )
    except AttributeError:
      other = int(other)
      if other < 0 or other > _upper[ nbits ]:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(_upper[ nbits ])}" )
      return _new_valid_bits( nbits, self._uint | other )

  def __ror__( self, other ):
    return self.__or__( other )

  def __xor__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of  '^' (xor) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( nbits, self._uint ^ other._uint )
    except AttributeError:
      other = int(other)
      if other < 0 or other > _upper[ self._nbits ]:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(_upper[ nbits ])}" )
      return _new_valid_bits( nbits, self._uint ^ other )

  def __rxor__( self, other ):
    return self.__xor__( other )

  def __floordiv__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '//' (div) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( nbits, (self._uint // other._uint) & _upper[nbits] )
    except AttributeError:
      other = int(other)
      if other < 0 or other > _upper[ self._nbits ]:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(_upper[ nbits ])}" )
      return _new_valid_bits( nbits, self._uint // other )

  def __rfloordiv__( self, other ):
    nbits = self._nbits
    other = int(other)
    up = _upper[ nbits ]
    if other < 0 or other > up:
      raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                        f"Suggestion: 0 <= x <= {hex(up)}" )
    return _new_valid_bits( nbits, other // self._uint )

  def __mod__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '%' (mod) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( nbits, (self._uint % other._uint) & _upper[nbits] )
    except AttributeError:
      other = int(other)
      if other < 0 or other > _upper[ nbits ]:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(_upper[ nbits ])}" )
      return _new_valid_bits( nbits, self._uint % other )

  def __rmod__( self, other ):
    nbits = self._nbits
    other = int(other)
    up = _upper[ nbits ]
    if other < 0 or other > up:
      raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                        f"Suggestion: 0 <= x <= {hex(up)}" )
    return _new_valid_bits( nbits, other % self._uint )

  def __invert__( self ):
    nbits = self._nbits
    return _new_valid_bits( nbits, ~self._uint & _upper[nbits] )

  def __lshift__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '<<' (lshift) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      uint = other._uint
      if uint >= nbits:
        return _new_valid_bits( self._nbits, 0 )
      return _new_valid_bits( nbits, (self._uint << uint) & _upper[nbits] )
    except AttributeError:
      other = int(other)
      if other < 0 or other > _upper[ nbits ]:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(_upper[ nbits ])}" )
      if other >= nbits:
        return _new_valid_bits( self._nbits, 0 )
      return _new_valid_bits( nbits, (self._uint << other) & _upper[nbits] )

  def __rshift__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '>>' (rshift) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( nbits, self._uint >> other._uint )
    except AttributeError:
      other = int(other)
      if other < 0 or other > _upper[ nbits ]:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(_upper[ nbits ])}" )
      return _new_valid_bits( nbits, self._uint >> other )

  def __eq__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '==' (eq) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( 1, self._uint == other._uint )
    except AttributeError:
      try:
        other = int(other)
      except:
        return _new_valid_bits( 1, 0 )

      if other < 0 or other > _upper[ nbits ]:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(_upper[ nbits ])}" )
      return _new_valid_bits( 1, self._uint == other )

  def __ne__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '!=' (ne) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( 1, self._uint != other._uint )
    except AttributeError:
      try:
        other = int(other)
      except:
        return _new_valid_bits( 1, 1 )

      if other < 0 or other > _upper[ nbits ]:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(_upper[ nbits ])}" )
      return _new_valid_bits( 1, self._uint != other )

  def __lt__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '<' (lt) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( 1, self._uint < other._uint )
    except AttributeError:
      other = int(other)
      if other < 0 or other > _upper[ self._nbits ]:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{self._nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(_upper[ self._nbits ])}" )
      return _new_valid_bits( 1, self._uint < other )

  def __le__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '<=' (le) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( 1, self._uint <= other._uint )
    except AttributeError:
      other = int(other)
      if other < 0 or other > _upper[ self._nbits ]:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{self._nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(_upper[ self._nbits ])}" )
      return _new_valid_bits( 1, self._uint <= other )

  def __gt__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '>' (gt) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( 1, self._uint > other._uint )
    except AttributeError:
      other = int(other)
      if other < 0 or other > _upper[ self._nbits ]:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{self._nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(_upper[ self._nbits ])}" )
      return _new_valid_bits( 1, self._uint > other )

  def __ge__( self, other ):
    nbits = self._nbits
    try:
      if other.nbits != nbits:
        raise ValueError( f"Operands of '>=' (ge) operation must have matching bitwidth, "\
                          f"but here Bits{nbits} != Bits{other.nbits}.\n" )
      return _new_valid_bits( 1, self._uint >= other._uint )
    except AttributeError:
      other = int(other)
      if other < 0 or other > _upper[ self._nbits ]:
        raise ValueError( f"Integer {hex(other)} is not a valid binop operand with Bits{self._nbits}!\n"
                          f"Suggestion: 0 <= x <= {hex(_upper[ self._nbits ])}" )
      return _new_valid_bits( 1, self._uint >= other )

  def __bool__( self ):
    return self._uint != 0

  def __int__( self ):
    return int(self._uint)

  def int( self ):
    if self._uint >> (self._nbits - 1):
      return -int(~self + 1)
    return self._uint

  def uint( self ):
    return self._uint

  def __index__( self ):
    return int(self._uint)

  def __hash__( self ):
    return hash((self._nbits, self._uint))

  # Print

  def __repr__(self):
    return "Bits{}(0x{})".format( self._nbits, "{:x}".format(int(self._uint)).zfill(((self._nbits-1)//4)+1) )

  def __str__(self):
    str = "{:x}".format(int(self._uint)).zfill(((self._nbits-1)//4)+1)
    return str

  def bin(self):
    str = "{:b}".format(int(self._uint)).zfill(self._nbits)
    return "0b"+str

  def oct( self ):
    str = "{:o}".format(int(self._uint)).zfill(((self._nbits-1)//3)+1)
    return "0o"+str

  def hex( self ):
    str = "{:x}".format(int(self._uint)).zfill(((self._nbits-1)//4)+1)
    return "0x"+str



"""
==========================================================================
helpers_test.py
==========================================================================
Test cases for helper functions

Author : Shunning Jiang
  Date : Nov 30, 2019
"""

from pymtl3.datatypes import *


def test_concat():
  x = concat( Bits128(0x1234567890abcdef1234567890abcdef),
              Bits252(0xfffffffff22222222222222222444441234567890abcdef1234567890abcdef) )
  assert x.nbits == 380
  assert x == mk_bits(380)(0x1234567890abcdef1234567890abcdeffffffffff22222222222222222444441234567890abcdef1234567890abcdef)

def test_zext():
  assert zext( Bits8(0xe), 24 ) == Bits24(0xe)

def test_sext():
  assert zext( Bits8(0xe), 24 ) == Bits24(0xe)

def test_clog2():
  assert clog2(7) == 3
  assert clog2(8) == 3
  assert clog2(9) == 4

def test_reduce_and():
  for i in range(255):
    assert not reduce_and( b8(i) )
  assert reduce_and( b8(255) )
  assert reduce_and( b512((2**512)-1) )

def test_reduce_or():
  for i in range(1, 256):
    assert reduce_or(i)
  assert not reduce_or( b8(0) )
  assert reduce_or( b512((2**512)-1) )

def test_reduce_xor():
  assert reduce_xor( b8(0b10101011) ) == 1
  assert reduce_xor( b8(0b10101010) ) == 0



#=========================================================================
# VerilogTranslationConfigs.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jan 28, 2020

from pymtl3.passes.backends.verilog import VerilogTranslationPass
from pymtl3.passes.PassConfigs import BasePassConfigs, Checker


class VerilogTranslationConfigs( BasePassConfigs ):

  Options = {
    # Translate the current instance?
    "enable" : False,

    # Give an explicit file name to the translated source
    "explicit_file_name" : "",

    # Give an explicit module name to the translated component
    "explicit_module_name" : "",

    # Remove module definition during synthesis
    # Note that this could be applied to non-top modules
    "no_synthesis" : False,

    # This module does not have clk pin during synthesis
    # Note that this could be applied to non-top modules
    # This option can only be enabled if no_synthesis is True
    "no_synthesis_no_clk" : False,

    # This module does not have reset pin during synthesis
    # Note that this could be applied to non-top modules
    # This option can only be enabled if no_synthesis is True
    "no_synthesis_no_reset" : False,
  }

  Checkers = {
    ("enable", "no_synthesis", "no_synthesis_no_clk", "no_synthesis_no_reset") :
    Checker( lambda v: isinstance( v, bool ), "expects a boolean" ),

    ("explicit_file_name", "explicit_module_name") :
    Checker( lambda v: isinstance(v, str), "expects a string" )
  }

  Pass = VerilogTranslationPass



#=========================================================================
# VerilogTranslationPass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 12, 2019
"""Translate a PyMTL component hierarhcy into SystemVerilog source code."""
import os

from pymtl3 import MetadataKey
from pymtl3.passes.BasePass import BasePass

from ..util.utility import verilog_cmp
from .VTranslator import VTranslator


class VerilogTranslationPass( BasePass ):
  """Translate a PyMTL component hierarchy into Verilog."""

  # Translation pass input pass data

  #: Enable translation on a component.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  enable                = MetadataKey(bool)

  #: Specify the filename of the translated source file.
  #:
  #: Type: ``str``; input
  #:
  #: Default value: value of ``explicit_module_name`` of the top level
  explicit_file_name    = MetadataKey(str)

  #: Specify the translated name of the component.
  #:
  #: Type: ``str``; input
  #:
  #: Default value: component class name concatenated with parameters
  explicit_module_name  = MetadataKey(str)

  #: Wrap component translation result within \`ifndef SYNTHESIS
  #: Enabling this option effectively removes the component from
  #: the hierarchy during logic synthesis.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  no_synthesis          = MetadataKey(bool)

  #: In the translated source, wrap the clk port connection of the
  #: component within \`ifndef SYNTHESIS.
  #: This option can only be enabled while ``no_synthesis`` is ``True``.
  #: Enabling this option effectively removes the clk pin of the component
  #: during logic synthesis.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  no_synthesis_no_clk   = MetadataKey(bool)

  #: In the translated source, wrap the reset port connection of the
  #: component within \`ifndef SYNTHESIS.
  #: This option can only be enabled while ``no_synthesis`` is ``True``.
  #: Enabling this option effectively removes the reset pin of the component
  #: during logic synthesis.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  no_synthesis_no_reset = MetadataKey(bool)

  # Translation pass output pass data

  #: An instance of :class:`TranslationConfigs` that contains the parsed options.
  #:
  #: Type: ``TranslationConfigs``; output
  translate_config      = MetadataKey()

  #: Whether or not the translated result is the same as the existing output file.
  #:
  #: Type: ``bool``; output
  is_same               = MetadataKey(bool)

  #: A reference of the translator called during translation.
  #:
  #: Type: output
  translator            = MetadataKey()

  #: Whether or not the component has been translated.
  #:
  #: Type: ``bool``; output
  translated            = MetadataKey(bool)

  #: Filename of the translation result.
  #:
  #: Type: ``str``; output
  translated_filename   = MetadataKey(str)

  #: Top level module name in the translation result.
  #:
  #: Type: ``str``; output
  translated_top_module = MetadataKey(str)

  def __call__( s, top ):
    """Translate a PyMTL component hierarhcy rooted at ``top``."""
    s.top = top
    s.translator = VTranslator( s.top )
    s.traverse_hierarchy( top )

  def get_translation_config( s ):
    from pymtl3.passes.backends.verilog.translation.VerilogTranslationConfigs import (
        VerilogTranslationConfigs,
    )
    return VerilogTranslationConfigs

  def gen_tr_cfgs( s, m ):
    tr_cfgs = {}

    def traverse( m ):
      nonlocal tr_cfgs
      tr_cfgs[m] = s.get_translation_config()( m )
      for _m in m.get_child_components(repr):
        traverse( _m )

    traverse( m )
    return tr_cfgs

  def traverse_hierarchy( s, m ):
    c = s.__class__

    if m.has_metadata( c.enable ) and m.get_metadata( c.enable ):
      m.set_metadata( c.translate_config, s.gen_tr_cfgs(m) )
      s.translator.translate( m, m.get_metadata( c.translate_config ) )

      module_name = s.translator._top_module_full_name

      if m.has_metadata( c.explicit_file_name ) and \
         m.get_metadata( c.explicit_file_name ):
        fname = m.get_metadata( c.explicit_file_name )
        if '.v' in fname:
          filename = fname.split('.v')[0]
        elif '.sv' in fname:
          filename = fname.split('.sv')[0]
        else:
          filename = fname

        output_file = filename + '.v'
        temporary_file = filename + '.v.tmp'

        # First write the file to a temporary file
        is_same = False
        with open( temporary_file, 'w' ) as output:
          output.write( s.translator.hierarchy.src )
          output.flush()
          os.fsync( output )
          output.close()

        # `is_same` is set if there exists a file that has the same filename as
        # `output_file`, and that file is the same as the temporary file
        if ( os.path.exists(output_file) ):
          is_same = verilog_cmp( temporary_file, output_file )

        # Rename the temporary file to the output file
        os.rename( temporary_file, output_file )

        # Expose some attributes about the translation process
        m.set_metadata( c.is_same,               is_same      )
        m.set_metadata( c.translator,            s.translator )
        m.set_metadata( c.translated,            True         )
        m.set_metadata( c.translated_filename,   output_file  )
        m.set_metadata( c.translated_top_module, module_name  )

      else:
        filename = f"{module_name}__pickled"

      output_file = filename + '.v'
      temporary_file = filename + '.v.tmp'

      # First write the file to a temporary file
      is_same = False
      with open( temporary_file, 'w' ) as output:
        output.write( s.translator.hierarchy.src )
        output.flush()
        os.fsync( output )
        output.close()

      # `is_same` is set if there exists a file that has the same filename as
      # `output_file`, and that file is the same as the temporary file
      if ( os.path.exists(output_file) ):
        is_same = verilog_cmp( temporary_file, output_file )

      # Rename the temporary file to the output file
      os.rename( temporary_file, output_file )

      # Expose some attributes about the translation process
      m.set_metadata( c.is_same,               is_same      )
      m.set_metadata( c.translator,            s.translator )
      m.set_metadata( c.translated,            True         )
      m.set_metadata( c.translated_filename,   output_file  )
      m.set_metadata( c.translated_top_module, module_name  )

    else:
      for child in m.get_child_components(repr):
        s.traverse_hierarchy( child )



#=========================================================================
# BehavioralRTLIRGenL3Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Oct 20, 2018
"""Provide L3 behavioral RTLIR generation pass."""

from pymtl3.datatypes import Bits, is_bitstruct_class, is_bitstruct_inst
from pymtl3.passes.rtlir.errors import PyMTLSyntaxError

from . import BehavioralRTLIR as bir
from .BehavioralRTLIRGenL2Pass import (
    BehavioralRTLIRGeneratorL2,
    BehavioralRTLIRGenL2Pass,
)


class BehavioralRTLIRGenL3Pass( BehavioralRTLIRGenL2Pass ):
  def get_rtlir_generator_class( s ):
    return BehavioralRTLIRGeneratorL3

class BehavioralRTLIRGeneratorL3( BehavioralRTLIRGeneratorL2 ):

  def visit_Call( s, node ):
    """Return behavioral RTLIR of a method call.

    At L3 we need to support the syntax of struct instantiation in upblks.
    This is achieved by function calls like `struct( 1, 2, 0 )`.
    """
    obj = s.get_call_obj( node )
    if is_bitstruct_class( obj ):
      fields = obj.__bitstruct_fields__
      nargs = len(node.args)
      nfields = len(fields.keys())
      if nargs == 0:
        # Infer the values of each field by inspecting the object constructed
        # with default arguments
        inst = obj()
        values = [s._datatype_to_bir(getattr(inst, field)) for field in fields.keys()]
      else:
        # Otherwise all fields of the struct must be present in the arguments
        if nargs != nfields:
          raise PyMTLSyntaxError( s.blk, node,
            f'BitStruct {obj.__name__} has {nfields} fields but {nargs} arguments are given!' )
        values = [s.visit(arg) for arg in node.args]

      ret = bir.StructInst( obj, values )
      ret.ast = node
      return ret

    else:
      return super().visit_Call( node )

  def _datatype_to_bir( s, instance ):
    if isinstance( instance, Bits ):
      return s._bits_to_bir( instance )
    elif is_bitstruct_inst( instance ):
      return s._struct_to_bir( instance )
    else:
      assert False, f"unrecognized datatype instance {instance}"

  def _struct_to_bir( s, instance ):
    struct_cls = instance.__class__
    fields = struct_cls.__bitstruct_fields__.keys()
    values = [s._datatype_to_bir(getattr(instance, field)) for field in fields]
    return bir.StructInst( struct_cls, values )

  def _bits_to_bir( s, instance ):
    nbits = instance.nbits
    value = int(instance)
    return bir.SizeCast( nbits, bir.Number( value ) )



#=========================================================================
# VStructuralTranslatorL4_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 29, 2019
"""Test the level 4 SystemVerilog structural translator."""

from collections import deque

import pytest

from pymtl3.passes.backends.verilog import VerilogPlaceholderPass
from pymtl3.passes.backends.verilog.util.test_utility import check_eq
from pymtl3.passes.backends.verilog.util.utility import verilog_reserved

from ....testcases import (
    CaseBits32ArrayConnectSubCompAttrComp,
    CaseBits32ConnectSubCompAttrComp,
    CaseChildExplicitModuleName,
    CaseConnectArraySubCompArrayStructIfcComp,
    CaseHeteroCompArrayComp,
)
from ..VStructuralTranslatorL4 import VStructuralTranslatorL4


def run_test( case, m ):
  m.elaborate()
  VStructuralTranslatorL4.is_verilog_reserved = lambda s, x: x in verilog_reserved
  tr = VStructuralTranslatorL4( m )
  tr._placeholder_pass = VerilogPlaceholderPass
  tr.clear( m )
  tr._rtlir_tr_unpacked_q = deque()
  tr.translate_structural( m )
  subcomps = tr.structural.decl_subcomps

  check_eq( subcomps[m], case.REF_COMP )

@pytest.mark.parametrize(
  'case', [
    CaseBits32ConnectSubCompAttrComp,
    CaseConnectArraySubCompArrayStructIfcComp,
    CaseBits32ArrayConnectSubCompAttrComp,
    CaseHeteroCompArrayComp,
  ]
)
def test_verilog_structural_L4( case ):
  run_test( case, case.DUT() )

def test_verilog_structural_L4_child_explicit_module_name():
  run_test( CaseChildExplicitModuleName, CaseChildExplicitModuleName.DUT() )



template = \
'''
// VT_INPUT_DELAY, VTB_OUTPUT_ASSERT_DELAY are timestamps relative to the rising edge.
`define VTB_INPUT_DELAY 1
`define VTB_OUTPUT_ASSERT_DELAY 3

// CYCLE_TIME and INTRA_CYCLE_TIME are duration of time.
`define CYCLE_TIME 4
`define INTRA_CYCLE_TIME (`VTB_OUTPUT_ASSERT_DELAY-`VTB_INPUT_DELAY)

`timescale 1ns/1ns

`define T({args_strs}) \\
        t({args_strs},`__LINE__)

// Tick one extra cycle upon an error.
`define VTB_TEST_FAIL(lineno, out, ref, port_name) \\
    $display("- Timestamp      : %0d (default unit: ns)", $time); \\
    $display("- Cycle number   : %0d (variable: cycle_count)", cycle_count); \\
    $display("- line number    : line %0d in {cases_file_name}", lineno); \\
    $display("- port name      : %s", port_name); \\
    $display("- expected value : 0x%x", ref); \\
    $display("- actual value   : 0x%x", out); \\
    $display(""); \\
    #(`CYCLE_TIME-`INTRA_CYCLE_TIME); \\
    cycle_count += 1; \\
    #`CYCLE_TIME; \\
    cycle_count += 1; \\
    $fatal;

`define CHECK(lineno, out, ref, port_name) \\
  if ((|(out ^ out)) == 1'b0) ; \\
  else begin \\
    $display(""); \\
    $display("The test bench received a value containing X/Z's! Please note"); \\
    $display("that the VTB is pessmistic about X's and you should make sure"); \\
    $display("all output ports of your DUT does not produce X's after reset."); \\
    `VTB_TEST_FAIL(lineno, out, ref, port_name) \\
  end \\
  if (out != ref) begin \\
    $display(""); \\
    $display("The test bench received an incorrect value!"); \\
    `VTB_TEST_FAIL(lineno, out, ref, port_name) \\
  end

module {harness_name};
  // convention
  logic clk;
  logic reset;
  integer cycle_count;

  {signal_decls};

  task t(
    {task_signal_decls},
    integer lineno
  );
  begin
    {task_assign_strs};
    #`INTRA_CYCLE_TIME;
    {task_check_strs};
    #(`CYCLE_TIME-`INTRA_CYCLE_TIME);
    cycle_count += 1;
  end
  endtask

  // use 25% clock cycle, so #1 for setup #2 for sim #1 for hold
  always #(`CYCLE_TIME/2) clk = ~clk;

  // DUT name
  // By default we use the translated name of the Verilog component. But you can change
  // that by defining the VTB_TOP_MODULE_NAME macro through the simulator command line
  // options (e.g., for VCS you can do +define+VTB_TOP_MODULE_NAME=YourTopModuleName).
`ifdef VTB_TOP_MODULE_NAME
  `VTB_TOP_MODULE_NAME DUT
`else
  {dut_name} DUT
`endif
  (
    {dut_clk_decl},
    {dut_reset_decl},
    {dut_signal_decls}
  );

  initial begin
    assert(0 <= `VTB_INPUT_DELAY)
      else $fatal("\\n=====\\n\\nVTB_INPUT_DELAY should >= 0\\n\\n=====\\n");

    assert(`VTB_INPUT_DELAY < `VTB_OUTPUT_ASSERT_DELAY)
      else $fatal("\\n=====\\n\\nVTB_OUTPUT_ASSERT_DELAY should be larger than VTB_INPUT_DELAY\\n\\n=====\\n");

    assert(`VTB_OUTPUT_ASSERT_DELAY <= `CYCLE_TIME)
      else $fatal("\\n=====\\n\\nVTB_OUTPUT_ASSERT_DELAY should be smaller than or equal to CYCLE_TIME\\n\\n=====\\n");

    cycle_count = 0;
    clk   = 1'b0; // NEED TO DO THIS TO HAVE FALLING EDGE AT TIME 0
    reset = 1'b1; // TODO reset active low/high
    #(`CYCLE_TIME/2);

    // Now we are talking
    #`VTB_INPUT_DELAY;
    #`CYCLE_TIME;
    cycle_count = 1;
    #`CYCLE_TIME;
    cycle_count = 2;
    // 2 cycles plus input delay
    reset = 1'b0;

    `include "{cases_file_name}"

    $display("");
    $display("  [ passed ]");
    $display("");

    // Tick one extra cycle for better waveform
    #`CYCLE_TIME;
    cycle_count += 1;
    $finish;
  end
endmodule
'''



"""
======================================================================
Crossbar.py
======================================================================
"""

from pymtl3 import *


class Crossbar( Component ):

  def construct( s, nports, dtype ):

    s.in_ = [ InPort  ( dtype )     for _ in range( nports ) ]
    s.out = [ OutPort ( dtype )     for _ in range( nports ) ]
    s.sel = [ InPort  ( clog2( nports ) ) for _ in range( nports ) ]

    @update
    def comb_logic():
      for i in range( nports ):
        s.out[i] @= s.in_[ s.sel[ i ] ]

  def line_trace( s ):
    in_str  = ' '.join( [ str(x) for x in s.in_ ] )
    sel_str = ' '.join( [ str(x) for x in s.sel ] )
    out_str = ' '.join( [ str(x) for x in s.out ] )
    return f"{in_str} ( {sel_str} ) {out_str}"



#=========================================================================
# BehavioralRTLIRTypeCheckL2Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 29, 2019
"""Provide L2 behavioral RTLIR type check pass."""

from pymtl3 import Bits32, MetadataKey
from pymtl3.passes.BasePass import PassMetadata
from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt

from . import BehavioralRTLIR as bir
from .BehavioralRTLIRGenL1Pass import BehavioralRTLIRGenL1Pass
from .BehavioralRTLIRTypeCheckL1Pass import (
    BehavioralRTLIRTypeCheckL1Pass,
    BehavioralRTLIRTypeCheckVisitorL1,
    BehavioralRTLIRTypeEnforcerL1,
)


class BehavioralRTLIRTypeCheckL2Pass( BehavioralRTLIRTypeCheckL1Pass ):

  # Pass metadata

  #: A dictionary that maps tmp variable names to its RTLIRType
  #:
  #: Type: ``dict``; output
  rtlir_tmpvars = MetadataKey()

  def get_visitor_class( s ):
    return BehavioralRTLIRTypeCheckVisitorL2

  def __call__( s, m ):
    """Perform type checking on all RTLIR in rtlir_upblks."""
    c = s.__class__

    rtlir_freevars = {}
    rtlir_accessed = set()
    rtlir_tmpvars  = {}

    m.set_metadata( BehavioralRTLIRTypeCheckL2Pass.rtlir_freevars, rtlir_freevars )
    m.set_metadata( BehavioralRTLIRTypeCheckL2Pass.rtlir_accessed, rtlir_accessed )
    m.set_metadata( BehavioralRTLIRTypeCheckL2Pass.rtlir_tmpvars,  rtlir_tmpvars )

    type_checker = s.get_visitor_class()(
      m,
      rtlir_freevars,
      rtlir_accessed,
      rtlir_tmpvars,
      s.tr_top.get_metadata( c.rtlir_getter ),
    )

    rtlir_upblks = m.get_metadata( BehavioralRTLIRGenL1Pass.rtlir_upblks )

    for blk in m.get_update_block_order():
      type_checker.enter( blk, rtlir_upblks[ blk ] )

#-------------------------------------------------------------------------
# Type checker
#-------------------------------------------------------------------------

class BehavioralRTLIRTypeCheckVisitorL2( BehavioralRTLIRTypeCheckVisitorL1 ):
  def __init__( s, component, freevars, accessed, tmpvars, rtlir_getter ):
    super().__init__(component, freevars, accessed, rtlir_getter)
    s.tmpvars = tmpvars
    s.tmpvars_is_explicit = {}
    s.loopvar_nbits = {}
    s.loopvar_is_explicit = {}
    s.BinOp_max_nbits = (bir.Add, bir.Sub, bir.Mult, bir.Div, bir.Mod, bir.Pow,
                         bir.BitAnd, bir.BitOr, bir.BitXor)
    s.BinOp_left_nbits = ( bir.ShiftLeft, bir.ShiftRightLogic )
    s.type_expect = {}
    s.enforcer = s.get_enforce_visitor()( component, s.tmpvars_is_explicit )
    lhs_types = ( rt.Port, rt.Wire, rt.NetWire, rt.NoneType )

    s.type_expect[ 'Assign' ] = (
      ( 'targets', lhs_types, 'lhs of assignment must be signal/tmpvar!' ),
      ( 'value',   rt.Signal, 'rhs of assignment should be signal/const!' ),
    )
    s.type_expect[ 'BinOp' ] = (
      ( 'left',    rt.Signal, 'lhs of binop should be signal/const!' ),
      ( 'right',   rt.Signal, 'rhs of binop should be signal/const!' ),
    )
    s.type_expect[ 'UnaryOp' ] = (
      ( 'operand', rt.Signal, 'unary op only applies to signals and consts!' ),
    )
    s.type_expect[ 'For' ] = (
      ( 'start',   rt.Const, 'the start of a for-loop must be a constant expression!' ),
      ( 'end',     rt.Const, 'the end of a for-loop must be a constant expression!' ),
      ( 'step',    rt.Const, 'the step of a for-loop must be a constant expression!' ),
    )
    s.type_expect[ 'If' ] = (
      ( 'cond',    rt.Signal, 'the condition of if must be a signal!' ),
    )
    s.type_expect[ 'IfExp' ] = {
      ( 'cond',    rt.Signal, 'the condition of if-exp must be a signal!' ),
      ( 'body',    rt.Signal, 'the body of if-exp must be a signal!' ),
      ( 'orelse',  rt.Signal, 'the else branch of if-exp must be a signal!' ),
    }

  def get_enforce_visitor( s ):
    return BehavioralRTLIRTypeEnforcerL2

  def eval_const_binop( s, l, op, r ):
    """Evaluate ( l op r ) and return the result as an integer."""
    assert type( l ) == int or isinstance( l, pymtl3_datatype.Bits )
    assert type( r ) == int or isinstance( r, pymtl3_datatype.Bits )
    op_dict = {
      # bir.And       : 'and', bir.Or    : 'or',
      bir.Add       : '+',   bir.Sub   : '-',  bir.Mult : '*',  bir.Div  : '/',
      bir.Mod       : '%',   bir.Pow   : '**',
      bir.ShiftLeft : '<<',  bir.ShiftRightLogic : '>>',
      bir.BitAnd    : '&',   bir.BitOr : '|',  bir.BitXor : '^',
    }
    _op = op_dict[ type( op ) ]
    return eval( f'l{_op}r' )

  def _visit_Assign_single_target( s, node, target, i ):
    rhs_type = node.value.Type
    lhs_type = target.Type

    if isinstance( target, bir.TmpVar ):
      tmpvar_id = (target.name, target.upblk_name)
      if lhs_type != rt.NoneType() and lhs_type.get_dtype() != rhs_type.get_dtype():
        raise PyMTLTypeError( s.blk, node.ast,
          f'conflicting type {rhs_type} for temporary variable {node.targets[i].name}(LHS target#{i+1} of {lhs_type})!' )

      # Creating a temporaray variable
      # Reminder that a temporary variable is essentially a wire. So we use
      # rt.Wire here instead of rt.NetWire
      target.Type = rt.Wire( rhs_type.get_dtype() )
      s.tmpvars[ tmpvar_id ] = rt.Wire( rhs_type.get_dtype() )
      s.tmpvars_is_explicit[ tmpvar_id ] = node.value._is_explicit

    else:
      # non-temporary assignment is an L1 thing
      super()._visit_Assign_single_target( node, target, i )

  def visit_Assign( s, node ):
    # RHS should have the same type as LHS
    for i, target in enumerate( node.targets ):
      s._visit_Assign_single_target( node, target, i )

    node.Type = None
    node._is_explicit = True

  def visit_If( s, node ):
    # Can the type of condition be cast into bool?
    dtype = node.cond.Type.get_dtype()
    if not rdt.Bool()( dtype ):
      raise PyMTLTypeError( s.blk, node.ast,
        'the condition of "if" cannot be converted to bool!' )
    node.Type = None
    node._is_explicit = True

  def visit_For( s, node ):
    try:
      if node.start._value < 0:
        raise PyMTLTypeError( s.blk, node.ast,
          'the start of for-loop must be non-negative!' )
      if node.end._value < 0:
        raise PyMTLTypeError( s.blk, node.ast,
          'the end of for-loop must be non-negative!' )
    except AttributeError:
      pass

    try:
      step = node.step._value
      if step == 0:
        raise PyMTLTypeError( s.blk, node.ast,
          'the step of for-loop cannot be zero!' )
    except AttributeError:
      raise PyMTLTypeError( s.blk, node.ast,
        'the step of for-loop must be a constant!' )

    if hasattr(node.start, '_value') and hasattr(node.end, '_value') and \
       hasattr(node.step, '_value'):
      loop_range = list(range(node.start._value, node.end._value, node.step._value))
      if len(loop_range) > 0:
        lvar_nbits = s._get_nbits_from_value(max(loop_range))
      else:
        lvar_nbits = s._get_nbits_from_value(max([node.start._value, node.end._value, node.step._value]))
      s.loopvar_is_explicit[node.var.name] = False
    else:
      lvar_nbits = max([x.Type.get_dtype().get_length() for x in [node.start, node.end, node.step]])
      s.loopvar_is_explicit[node.var.name] = True

    # context_type = rt.NetWire(rdt.Vector(lvar_nbits))
    s.loopvar_nbits[node.var.name] = lvar_nbits
    # s.enforcer.enter(s.blk, context_type, node.start)
    # s.enforcer.enter(s.blk, context_type, node.end)
    # s.enforcer.enter(s.blk, context_type, node.step)

    for stmt in node.body:
      s.visit( stmt )

    del s.loopvar_nbits[node.var.name]

    node.Type = None
    node._is_explicit = True

  def visit_LoopVar( s, node ):
    node.Type = rt.Const( rdt.Vector( s.loopvar_nbits[node.name] ), None )
    node._is_explicit = s.loopvar_is_explicit[node.name]

  def visit_TmpVar( s, node ):
    tmpvar_id = (node.name, node.upblk_name)
    if tmpvar_id not in s.tmpvars:
      # This tmpvar is being created. Later when it is used, its type can
      # be read from the tmpvar type environment.
      node.Type = rt.NoneType()
      node._is_explicit = True

    else:
      node.Type = s.tmpvars[ tmpvar_id ]
      node._is_explicit = s.tmpvars_is_explicit[ tmpvar_id ]

  def visit_IfExp( s, node ):
    # Can the type of condition be cast into bool?
    if not rdt.Bool()( node.cond.Type.get_dtype() ):
      raise PyMTLTypeError( s.blk, node.ast,
        'the condition of "if-exp" cannot be converted to bool!' )

    # body and orelse must have the same type
    # if node.body.Type != node.orelse.Type:
    if not node.body.Type.get_dtype()( node.orelse.Type.get_dtype() ):
      raise PyMTLTypeError( s.blk, node.ast,
        'the body and orelse of "if-exp" must have the same type!' )

    lhs_dtype, rhs_dtype = node.body.Type.get_dtype(), node.orelse.Type.get_dtype()
    lhs_is_vector = isinstance(lhs_dtype, rdt.Vector)
    rhs_is_vector = isinstance(rhs_dtype, rdt.Vector)
    lhs_nbits, rhs_nbits = lhs_dtype.get_length(), rhs_dtype.get_length()

    # Unify body and orelse if both are rdt.Vector
    if lhs_is_vector and rhs_is_vector and lhs_nbits != rhs_nbits:
      is_lhs_inferred = not node.body._is_explicit
      is_rhs_inferred = not node.orelse._is_explicit

      # Both sides are explicit
      if not is_lhs_inferred and not is_rhs_inferred:
        raise PyMTLTypeError( s.blk, node.ast,
          f'the body and orelse of "if-exp" have different bitwidth {lhs_nbits} vs {rhs_nbits}!' )

      # Both sides are implicit
      elif is_lhs_inferred and is_rhs_inferred:
        if lhs_nbits >= rhs_nbits:
          target_nbits = lhs_nbits
          op = node.body
        else:
          target_nbits = rhs_nbits
          op = node.orelse
        context = rt.NetWire(rdt.Vector(target_nbits))
        s.enforcer.enter( s.blk, context, op )

      else:
        # One side is explicit and the other implicit
        if is_lhs_inferred:
          exp_str, imp_str = "or-else", "body"
          context, op, explicit, implicit = node.orelse.Type, node.body, rhs_nbits, lhs_nbits
        else:
          exp_str, imp_str = "body", "or-else"
          context, op, explicit, implicit = node.body.Type, node.orelse, lhs_nbits, rhs_nbits
        if explicit < implicit:
          raise PyMTLTypeError( s.blk, node.ast,
              f"The {exp_str} side of if-exp has {explicit} bits but "
              f"the {imp_str} side requires more bits ({implicit})!" )
        s.enforcer.enter( s.blk, context, op )

    node.Type = node.body.Type
    node._is_explicit = node.body._is_explicit or node.orelse._is_explicit

  def visit_UnaryOp( s, node ):
    # if isinstance( node.op, bir.Not ):
    #   dtype = node.operand.Type.get_dtype()
    #   if not rdt.Bool()( dtype ):
    #     raise PyMTLTypeError( s.blk, node.ast,
    #       'the operand of "Logic-not" cannot be cast to bool!' )
    #   if dtype.get_length() != 1:
    #     raise PyMTLTypeError( s.blk, node.ast,
    #       'the operand of "Logic-not" is not a single bit!' )
    #   node.Type = rt.NetWire( rdt.Bool() )
    # else:
    node.Type = node.operand.Type
    node._is_explicit = node.operand._is_explicit
    if hasattr( node.operand, '_value' ):
      opmap = {
          # bir.Not    : 'not',
          bir.Invert : '~',
          bir.UAdd   : '+',
          bir.USub   : '-',
      }
      try:
        op = opmap[node.op.__class__]
        operand = node.operand._value
        node._value = eval(f"{op}{operand}")
      except:
        pass

  # def visit_BoolOp( s, node ):
  #   max_nbits = -1
  #   for value in node.values:
  #     dtype = value.Type.get_dtype()
  #     if not isinstance(value.Type, rt.Signal) or not rdt.Bool()(dtype):
  #       raise PyMTLTypeError( s.blk, node.ast,
  #         f"{value} of {value.Type} cannot be cast into bool!")
  #     if dtype.get_length() > max_nbits:
  #       max_nbits = dtype.get_length()
  #   node.Type = rt.NetWire( rdt.Bool(max_nbits) )
  #   node._is_explicit = True

  def visit_BinOp( s, node ):
    op = node.op
    l_type = node.left.Type.get_dtype()
    r_type = node.right.Type.get_dtype()
    l_explicit, r_explicit = node.left._is_explicit, node.right._is_explicit
    if not( rdt.Vector(1)( l_type ) and rdt.Vector(1)( r_type ) ):
      raise PyMTLTypeError( s.blk, node.ast,
        f"both sides of {op.__class__.__name__} should be of vector type!" )

    l_nbits = l_type.get_length()
    r_nbits = r_type.get_length()

    # Enforcing Verilog bitwidth inference rules
    res_nbits = 0
    if isinstance( op, s.BinOp_max_nbits ):
      if (not l_explicit and r_explicit) or (l_explicit and not r_explicit):

        context, op, explicit, implicit = node.left.Type, node.right, l_nbits, r_nbits
        if not l_explicit:
          context, op, explicit, implicit = node.right.Type, node.left, r_nbits, l_nbits
        # Check if any implicit truncation happens
        if explicit < implicit:
          raise PyMTLTypeError( s.blk, node.ast,
              f"The explicitly sized side of operation has {explicit} bits but "
              f"the integer literal requires more bits ({implicit})!" )
        s.enforcer.enter( s.blk, context, op )

      elif not l_explicit and not r_explicit:
        # Both sides are implicit
        if l_nbits >= r_nbits:
          target_nbits = l_nbits
          op = node.right
        else:
          target_nbits = r_nbits
          op = node.left
        context = rt.NetWire(rdt.Vector(target_nbits))
        s.enforcer.enter( s.blk, context, op )

      else:
        # Both sides are explicit
        if not isinstance( op, s.BinOp_left_nbits ) and l_type != r_type:
          raise PyMTLTypeError( s.blk, node.ast,
            f"LHS and RHS of {op.__class__.__name__} should have the same type ({l_type} vs {r_type})!" )

      res_nbits = max( l_nbits, r_nbits )
      node._is_explicit = l_explicit or r_explicit

    elif isinstance( op, s.BinOp_left_nbits ):
      res_nbits = l_nbits
      node._is_explicit = l_explicit

    else:
      raise Exception( 'RTLIRTypeCheck internal error: unrecognized op!' )

    try:
      # Both sides are constant expressions
      l_val = node.left._value
      r_val = node.right._value
      node._value = s.eval_const_binop( l_val, node.op, r_val )
      node.Type = s.rtlir_getter.get_rtlir( node._value )
      assert isinstance( node.Type, rt.Const )
    except AttributeError:
      # Both sides are constant but the value cannot be determined statically
      if isinstance(node.left.Type, rt.Const) and isinstance(node.right.Type, rt.Const):
        node.Type = rt.Const( rdt.Vector( res_nbits ), None )
      # Variable
      else:
        node.Type = rt.NetWire( rdt.Vector( res_nbits ) )

  def visit_Compare( s, node ):
    l_type = node.left.Type.get_dtype()
    r_type = node.right.Type.get_dtype()
    l_explicit, r_explicit = node.left._is_explicit, node.right._is_explicit
    l_nbits, r_nbits = l_type.get_length(), r_type.get_length()

    if l_explicit and r_explicit:
      if l_type != r_type:
        raise PyMTLTypeError( s.blk, node.ast,
          f"LHS and RHS of {node.op.__class__.__name__} have different types ({l_type} vs {r_type})!" )

    elif not l_explicit and not r_explicit:
      if l_nbits >= r_nbits:
        target_nbits = l_nbits
        op = node.right
      else:
        target_nbits = r_nbits
        op = node.left
      context = rt.NetWire(rdt.Vector(target_nbits))
      s.enforcer.enter( s.blk, context, op )

    else:
      context, op, explicit, implicit = node.left.Type, node.right, l_nbits, r_nbits
      if not l_explicit:
        context, op, explicit, implicit = node.right.Type, node.left, r_nbits, l_nbits
      # Check if any implicit truncation happens
      if explicit < implicit:
        raise PyMTLTypeError( s.blk, node.ast,
            f"The explicitly sized side of comparison has {explicit} bits but "
            f"the integer literal requires more bits ({implicit}) to hold!" )
      s.enforcer.enter( s.blk, context, op )

    node.Type = rt.NetWire( rdt.Bool() )
    node._is_explicit = True

#-------------------------------------------------------------------------
# Enforce types for all terms whose types are inferred (implicit)
#-------------------------------------------------------------------------

class BehavioralRTLIRTypeEnforcerL2( BehavioralRTLIRTypeEnforcerL1 ):
  
  def __init__( s, component, tmpvars_is_explicit ):
    s.BinOp_max_nbits = (bir.Add, bir.Sub, bir.Mult, bir.Div, bir.Mod, bir.Pow,
                         bir.BitAnd, bir.BitOr, bir.BitXor)
    s.BinOp_left_nbits = ( bir.ShiftLeft, bir.ShiftRightLogic )
    s.tmpvars_is_explicit = tmpvars_is_explicit
    super().__init__( component )

  def visit_TmpVar( s, node ):
    tmpvar_id = (node.name, node.upblk_name)
    if not s.tmpvars_is_explicit[tmpvar_id]:
      s.mutate_datatype( node, f'tmpvar {node.name}' )

  def visit_LoopVar( s, node ):
    s.mutate_datatype( node, f'loop variable {node.name}' )

  def visit_IfExp( s, node ):
    # The fact that we are here means both body and orelse should be
    # implicit. We need to perform two separate mutations on expressions
    # located at body and orelse.
    s.visit( node.body )
    s.visit( node.orelse )

    # After both child have been successfully updated, we need to update the
    # IfExp node itself.
    s.mutate_datatype( node, 'if-expression' )



#=========================================================================
# TranslationImport_stdlib_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 3, 2019
"""Test stdlib RTL components with SystemVerilog translation and import.

We reuse all the test cases from stdlib test files. To achieve this I
overwrite the original reference to the test function used in stdlib test
files and add my own test function (the `run_test`s). However we need
to make sure the orignal reference is not lost and is restored after
finishing each test (no matter it fails or passes).
"""

from pymtl3.passes.PassGroups import DefaultPassGroup

from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.basic_rtl.test.arbiters_test import test_rr_arb_4 as _rr_arb_4
from pymtl3.stdlib.basic_rtl.test.arbiters_test import test_rr_arb_en_4 as _rr_arb_en_4
from pymtl3.stdlib.basic_rtl.test.crossbars_test import test_crossbar3 as _crossbar3
from pymtl3.stdlib.basic_rtl.test.encoders_test import (
    test_encoder_5_directed as _encoder5,
)
from pymtl3.stdlib.stream.test.queues_test import test_normal1_simple as _n1
from pymtl3.stdlib.stream.test.queues_test import test_normal2_simple as _n2
from pymtl3.stdlib.stream.test.queues_test import test_bypass1_simple as _b1
from pymtl3.stdlib.stream.test.queues_test import test_pipe1_simple as _p1
from pymtl3.stdlib.test_utils import TestVectorSimulator
from pymtl3.stdlib.test_utils import run_sim as stdlib_run_sim

from ..VerilogTranslationImportPass import VerilogTranslationImportPass


def local_do_test( _m ):
  _m.elaborate()
  # Mark component `_m` as to be translated and imported
  _m.set_metadata( VerilogTranslationImportPass.enable, True )
  m = VerilogTranslationImportPass()( _m )
  sim = TestVectorSimulator( m, _m._tvs, _m._tv_in, _m._tv_out )
  sim.run_test()

def test_arbiter_rr_arb_4( do_test ):
  # Customized test function
  def run_test( m, tvs ):
    def tv_in( model, tv ):
      model.reqs @= tv[0]
    def tv_out( model, tv ):
      assert model.grants == tv[1]
    m._tvs = tvs
    m._tv_in, m._tv_out = tv_in, tv_out
    do_test( m )

  # Swap in the customized test method
  test_func = _rr_arb_4
  _run_test = test_func.__globals__['run_test']
  test_func.__globals__['run_test'] = run_test
  try:
    test_func()
  finally:
    # Restore the original method regardless of the test result
    test_func.__globals__['run_test'] = _run_test

def test_arbiter_rr_arb_en_4( do_test ):
  def run_test( m, tvs ):
    def tv_in( model, tv ):
      model.en   @= tv[0]
      model.reqs @= tv[1]
    def tv_out( model, tv ):
      assert model.grants == tv[2]
    m._tvs = tvs
    m._tv_in, m._tv_out = tv_in, tv_out
    do_test( m )

  test_func = _rr_arb_en_4
  _run_test = test_func.__globals__['run_en_test']
  test_func.__globals__['run_en_test'] = run_test
  try:
    test_func( None )
  finally:
    test_func.__globals__['run_en_test'] = _run_test

def test_crossbar3( do_test ):
  def run_test( m, tvs ):
    def tv_in( model, tv ):
      n = len( model.in_ )
      for i in range(n):
        model.in_[i] @= tv[i]
        model.sel[i] @= tv[n+i]
    def tv_out( model, tv ):
      n = len( model.in_ )
      for i in range(n):
        assert model.out[i] == tv[n*2+i]
    m._tvs = tvs
    m._tv_in, m._tv_out = tv_in, tv_out
    do_test( m )

  test_func = _crossbar3
  _run_test = test_func.__globals__['run_test_crossbar']
  test_func.__globals__['run_test_crossbar'] = run_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test_crossbar'] = _run_test

def test_encoder_5_directed( do_test ):
  def run_test( m, tvs ):
    def tv_in( model, tv ):
      model.in_ @= tv[0]
    def tv_out( model, tv ):
      assert model.out == tv[1]
    m._tvs = tvs
    m._tv_in, m._tv_out = tv_in, tv_out
    do_test( m )

  test_func = _encoder5
  _run_test = test_func.__globals__['run_test']
  test_func.__globals__['run_test'] = run_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test'] = _run_test

#-------------------------------------------------------------------------
# Queue tests
#-------------------------------------------------------------------------

# Swap run_test_queue with modified version that calls the local run_sim
def _run_queue_test( test_func ):

  def _run_sim( th, cmdline_opts, duts ):
    th.elaborate()
    dut_objs = []

    for dut in duts:
      dut_objs.append( eval(f'th.{dut}') )

    for obj in dut_objs:
      obj.set_metadata( VerilogTranslationImportPass.enable, True )

    th = VerilogTranslationImportPass()( th )

    th.apply( DefaultPassGroup() )
    th.sim_reset()

    while not th.done() and th.sim_cycle_count() < 100:
      th.sim_tick()

    assert th.sim_cycle_count() < 100

  original_run_sim = test_func.__globals__['run_sim']
  test_func.__globals__['run_sim'] = _run_sim
  try:
    test_func( None )
  finally:
    test_func.__globals__['run_sim'] = original_run_sim

def test_normal1_simple():
  _run_queue_test( _n1 )

def test_normal2_simple():
  _run_queue_test( _n2 )

def test_bypass1_simple():
  _run_queue_test( _b1 )

def test_pipe1_simple():
  _run_queue_test( _p1 )



#=========================================================================
# YosysStructuralTranslatorL1_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 13, 2019
"""Test the level 1 yosys-SystemVerilog structural translator."""

import pytest

from pymtl3.passes.backends.verilog.util.test_utility import check_eq
from pymtl3.passes.backends.verilog.util.utility import verilog_reserved

from ....testcases import (
    CaseBitSelOverBitSelComp,
    CaseBitSelOverPartSelComp,
    CaseConnectBitsConstToOutComp,
    CaseConnectBitSelToOutComp,
    CaseConnectConstToOutComp,
    CaseConnectInToWireComp,
    CaseConnectSliceToOutComp,
    CasePartSelOverBitSelComp,
    CasePartSelOverPartSelComp,
)
from ..YosysStructuralTranslatorL1 import YosysStructuralTranslatorL1


def run_test( case, m ):
  m.elaborate()
  YosysStructuralTranslatorL1.is_verilog_reserved = lambda s, x: x in verilog_reserved
  tr = YosysStructuralTranslatorL1( m )
  tr.clear( m )
  tr.translate_structural( m )

  ports = tr.structural.decl_ports[m]
  wires = tr.structural.decl_wires[m]
  conns = tr.structural.connections[m]

  check_eq( ports["port_decls"],  case.REF_PORTS_PORT )
  check_eq( ports["wire_decls"],  case.REF_PORTS_WIRE )
  check_eq( ports["connections"], case.REF_PORTS_CONN )
  check_eq( wires, case.REF_WIRE )
  check_eq( conns, case.REF_CONN )

@pytest.mark.parametrize(
  'case', [
    CaseConnectInToWireComp,
    CaseConnectBitsConstToOutComp,
    CaseConnectConstToOutComp,
    CaseConnectBitSelToOutComp,
    CaseConnectSliceToOutComp,
    CaseBitSelOverBitSelComp,
    CaseBitSelOverPartSelComp,
    CasePartSelOverBitSelComp,
    CasePartSelOverPartSelComp,
  ]
)
def test_yosys_structural_L1( case ):
  run_test( case, case.DUT() )



from .arbiters import RoundRobinArbiter, RoundRobinArbiterEn
from .arithmetics import (
    Adder,
    And,
    Demux,
    EqComparator,
    Incrementer,
    LEComparator,
    LeftLogicalShifter,
    LTComparator,
    Mux,
    RightLogicalShifter,
    Subtractor,
    ZeroComparator,
)
from .crossbars import Crossbar
from .encoders import Encoder
from .register_files import RegisterFile, RegisterFileRst
from .registers import Reg, RegEn, RegEnRst, RegRst



'''
==========================================================================
connect_bits2bitstruct_test.py
==========================================================================
Unit tests for connect_bits2bitstruct.

Author : Yanghui Ou
  Date : Feb 24, 2020
'''
from pymtl3 import *

from ..connect_bits2bitstruct import connect_bits2bitstruct

#-------------------------------------------------------------------------
# Bitstructs for unit tests
#-------------------------------------------------------------------------

@bitstruct
class Point:
  x : Bits8
  y : Bits4

@bitstruct
class Nested:
  z    : Bits6
  pt   : Point
  flag : Bits1

#-------------------------------------------------------------------------
# Unit test for imatmul casting
#-------------------------------------------------------------------------

def test_imatmul_bits2bitstruct_simple():

  class Bits2Bitstruct( Component ):
    def construct( s ):
      s.pt_bits      = InPort ( Bits12 )
      s.pt_bitstruct = OutPort( Point  )
      @update
      def up_2():
        s.pt_bitstruct @= s.pt_bits

  a = Bits2Bitstruct()
  a.elaborate()
  a.apply( DefaultPassGroup() )
  a.pt_bits @= 0xeda
  a.sim_eval_combinational()
  assert a.pt_bitstruct.x == 0xed
  assert a.pt_bitstruct.y == 0xa

def test_imatmul_bitstruct2bits_simple():

  class Bitstruct2Bits( Component ):
    def construct( s ):
      s.pt_bitstruct = InPort ( Point  )
      s.pt_bits      = OutPort( Bits12 )
      @update
      def up_2():
        s.pt_bits @= s.pt_bitstruct

  b = Bitstruct2Bits()
  b.elaborate()
  b.apply( DefaultPassGroup() )
  b.pt_bitstruct.x @= 0xed
  b.pt_bitstruct.y @= 0xa
  b.sim_eval_combinational()
  assert b.pt_bits == 0xeda

def test_imatmul_bits2bitstruct_nested():

  class Bits2BitstructNested( Component ):
    def construct( s ):
      s.pt_bits      = InPort ( Bits19 )
      s.pt_bitstruct = OutPort( Nested )
      @update
      def up_2():
        s.pt_bitstruct @= s.pt_bits

  a = Bits2BitstructNested()
  a.elaborate()
  a.apply( DefaultPassGroup() )
  a.pt_bits @= concat( b6(0x3f), b12(0xeda), b1(1) )
  a.sim_eval_combinational()
  assert a.pt_bitstruct.z    == 0x3f
  assert a.pt_bitstruct.pt.x == 0xed
  assert a.pt_bitstruct.pt.y == 0xa
  assert a.pt_bitstruct.flag == 1

def test_imatmul_bitstruct2bits_nested():

  class Bitstruct2BitsNested( Component ):
    def construct( s ):
      s.pt_bitstruct = InPort ( Nested )
      s.pt_bits      = OutPort( Bits19 )
      @update
      def up_2():
        s.pt_bits @= s.pt_bitstruct

  b = Bitstruct2BitsNested()
  b.elaborate()
  b.apply( DefaultPassGroup() )
  b.pt_bitstruct.z    @= 0x3f
  b.pt_bitstruct.pt.x @= 0xed
  b.pt_bitstruct.pt.y @= 0xa
  b.pt_bitstruct.flag @= 0x1
  b.sim_eval_combinational()
  assert b.pt_bits == concat( b6(0x3f), b12(0xeda), b1(1) )

#-------------------------------------------------------------------------
# Components for connect tests
#-------------------------------------------------------------------------

class Bits2Bitstruct( Component ):
  def construct( s ):
    s.pt_bits      = InPort ( Bits12 )
    s.pt_bitstruct = OutPort( Point  )
    connect_bits2bitstruct( s.pt_bits, s.pt_bitstruct )

class Bitstruct2Bits( Component ):
  def construct( s ):
    s.pt_bitstruct = InPort ( Point  )
    s.pt_bits      = OutPort( Bits12 )
    @update
    def up_2():
      s.pt_bits @= s.pt_bitstruct

class Bits2BitstructNested( Component ):
  def construct( s ):
    s.pt_bits      = InPort ( Bits19 )
    s.pt_bitstruct = OutPort( Nested )
    @update
    def up_2():
      s.pt_bitstruct @= s.pt_bits

class Bitstruct2BitsNested( Component ):
  def construct( s ):
    s.pt_bitstruct = InPort ( Nested )
    s.pt_bits      = OutPort( Bits19 )
    @update
    def up_2():
      s.pt_bits @= s.pt_bitstruct

#-------------------------------------------------------------------------
# Unit test fos connect_bits2bitstruct
#-------------------------------------------------------------------------

def test_connect_bits2bitstruct_simple():

  class Bits2Bitstruct( Component ):
    def construct( s ):
      s.pt_bits      = InPort ( Bits12 )
      s.pt_bitstruct = OutPort( Point  )
      connect_bits2bitstruct( s.pt_bits, s.pt_bitstruct )

  a = Bits2Bitstruct()
  a.elaborate()
  a.apply( DefaultPassGroup() )
  a.pt_bits @= 0xeda
  a.sim_eval_combinational()
  assert a.pt_bitstruct.x == 0xed
  assert a.pt_bitstruct.y == 0xa

def test_connect_bitstruct2bits_simple():

  class Bitstruct2Bits( Component ):
    def construct( s ):
      s.pt_bitstruct = InPort ( Point  )
      s.pt_bits      = OutPort( Bits12 )
      connect_bits2bitstruct( s.pt_bitstruct, s.pt_bits )

  b = Bitstruct2Bits()
  b.elaborate()
  b.apply( DefaultPassGroup() )
  b.pt_bitstruct.x @= 0xed
  b.pt_bitstruct.y @= 0xa
  b.sim_eval_combinational()
  assert b.pt_bits == 0xeda

def test_connect_bits2bitstruct_nested():

  class Bits2BitstructNested( Component ):
    def construct( s ):
      s.pt_bits      = InPort ( Bits19 )
      s.pt_bitstruct = OutPort( Nested )
      connect_bits2bitstruct( s.pt_bits, s.pt_bitstruct )

  a = Bits2BitstructNested()
  a.elaborate()
  a.apply( DefaultPassGroup() )
  a.pt_bits @= concat( b6(0x3f), b12(0xeda), b1(1) )
  a.sim_eval_combinational()
  assert a.pt_bitstruct.z    == 0x3f
  assert a.pt_bitstruct.pt.x == 0xed
  assert a.pt_bitstruct.pt.y == 0xa
  assert a.pt_bitstruct.flag == 1

def test_connect_bitstruct2bits_nested():

  class Bitstruct2BitsNested( Component ):
    def construct( s ):
      s.pt_bitstruct = InPort ( Nested )
      s.pt_bits      = OutPort( Bits19 )
      connect_bits2bitstruct( s.pt_bitstruct, s.pt_bits )

  b = Bitstruct2BitsNested()
  b.elaborate()
  b.apply( DefaultPassGroup() )
  b.pt_bitstruct.z    @= b6( 0x3f )
  b.pt_bitstruct.pt.x @= b8( 0xed )
  b.pt_bitstruct.pt.y @= b4( 0xa  )
  b.pt_bitstruct.flag @= b1( 0x1  )
  b.sim_eval_combinational()
  assert b.pt_bits == concat( b6(0x3f), b12(0xeda), b1(1) )



"""
========================================================================
Slicing_test.py
========================================================================

Author : Shunning Jiang
Date   : Aug 23, 2018
"""
from pymtl3.datatypes import Bits2, Bits4, Bits14, Bits16, Bits24, Bits32
from pymtl3.dsl.ComponentLevel1 import update
from pymtl3.dsl.ComponentLevel3 import ComponentLevel3, connect
from pymtl3.dsl.Connectable import Wire
from pymtl3.dsl.errors import MultiWriterError, NoWriterError

from .sim_utils import simple_sim_pass

# Shunning: Note that we now have @= as a hook for LHS, we can safely
# assigning constants
# >>> s.A[0:16] @= 0xff
# instead of s.A[0:16] @= Bits16( 0xff )

def _test_model( cls ):
  A = cls()
  A.elaborate()
  simple_sim_pass( A, 0x123 )

  for i in range(10):
    A.tick()

# write two disjoint slices
def test_write_two_disjoint_slices():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_0_16():
        s.A[0:16] @= Bits16( 0xff )

      @update
      def up_wr_16_30():
        s.A[16:30] @= Bits14( 0xff )

      @update
      def up_rd_12_30():
        assert s.A[12:30] == 0xff0

  _test_model( Top )

# write two disjoint slices, but one slice is not read at all
def test_write_two_disjoint_slices_no_reader():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_0_16():
        s.A[0:16] @= Bits16( 0xff )

      @update
      def up_wr_16_30():
        s.A[16:30] @= Bits14( 0xff )

      @update
      def up_rd_17_30():
        assert s.A[16:30] == 0xff

  m = Top()
  m.elaborate()
  simple_sim_pass( m, 0x123 )

  # assert len(m._all_constraints) == 1
  # x, y = list(m._all_constraints)[0]

  # assert  m._all_id_upblk[x].__name__ == "up_wr_16_30" and \
          # m._all_id_upblk[y].__name__ == "up_rd_17_30" # only one constraint


# write two overlapping slices
def test_write_two_overlapping_slices():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_0_24():
        s.A[0:24] @= Bits24( 0xff )

      @update
      def up_wr_8_32():
        s.A[8:32] @= Bits24( 0xff )

      @update
      def up_rd_A():
        x = s.A

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# write two slices and a single bit
def test_write_two_slices_and_bit():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_0_16():
        s.A[0:16] @= 0xff

      @update
      def up_wr_16_30():
        s.A[16:30] @= Bits14( 0xff )

      @update
      def up_wr_30_31():
        s.A[30] @= 1

      @update
      def up_rd_A():
        print(s.A[0:17])

  m = Top()
  m.elaborate()
  simple_sim_pass( m, 0x123 )

  # assert len(m._all_constraints) == 2
  # _, x = list(m._all_constraints)[0]
  # _, y = list(m._all_constraints)[1]

  # two constraints are: up_wr_0_16 < up_rd_A and up_wr_16_30 < up_rd_A
  # assert  m._all_id_upblk[x].__name__ == "up_rd_A" and \
          # m._all_id_upblk[y].__name__ == "up_rd_A"

# write a slice and a single bit, but they are overlapped
def test_write_slices_and_bit_overlapped():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_0_16():
        s.A[0:16] @= 0xff

      @update
      def up_wr_15():
        s.A[15] @= 1

      @update
      def up_rd_A():
        print(s.A[0:17])

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# write a slice and there are two reader
def test_multiple_readers():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_8_24():
        s.A[8:24] @= Bits16( 0x1234 )

      @update
      def up_rd_0_12():
        assert s.A[0:12] == 0x400

      @update
      def up_rd_bunch():
        assert s.A[23] == 0
        assert s.A[22] == 0
        assert s.A[21] == 0
        assert s.A[20] == 1
        assert s.A[19] == 0
        assert s.A[18] == 0
        assert s.A[17] == 1
        assert s.A[16] == 0

  _test_model( Top )

# 1. WR A[s], RD A    (A[s] (=) A, SAME AS data struct)
#    WR A[s], WR A    (detect 2-writer conflict, SAME AS data struct)
#    WR A[s], RD A[t] (A[s] (=) A[t] if s intersects t)
#    WR A[s], WR A[t] (detect 2-writer conflict if s intersects t)

# 2. WR A   , RD A[s] (A[s] (=) A, SAME AS data struct)

# 3. WR A[s], A   |=y, RD y (mark A as writer in net {A,y}, SAME AS data struct)
#    WR A[s], A   |=y, WR y (detect 2-writer conflict, SAME AS data struct)
#    WR A[s], A[t]|=y, RD y (mark A[t] as writer in net {A[t],y} if s intersects t)
#    WR A[s], A[t]|=y, WR y (detect 2-writer conflict if s intersects t)

# 4. WR A   , A[s]|=y, RD y (mark A[s] as writer in net {A[s],y}, SAME AS data struct)
#    WR A   , A[s]|=y, WR y (detect 2-writer conflict, SAME AS data struct)

# 5. WR x, x|=A[s], RD A    (A[s] (=) A, SAME AS data struct)
#    WR x, x|=A[s], RD A[t] (A[s] (=) A[t] if s intersects t)

# 6. WR x, x|=A   , RD A[s] (A[s] (=) A, SAME AS data struct)

# 7. WR x, x|=A[s], A   |=y, RD y (mark A as writer and implicit constraint)
#    WR x, x|=A[s], A   |=y, WR y (detect 2-writer conflict)
#    WR x, x|=A[s], A[t]|=y, RD y (mark A[t] as writer and implicit constraint if s intersects t)
#    WR x, x|=A[s], A[t]|=y, WR y (detect 2-writer conflict if s intersects t)

# 8. WR x, x|=A   , A[s]|=y, RD y (mark A[s] as writer in net {A[s],y}, SAME AS data struct)

# --------------------------------------------------------------------------

# RD A[s]
#  - WR A          (A[s] (=) A,                          SAME AS data struct)
#  - WR A[t]       (A[s] (=) A[t]                          if s intersects t)
#  - A   |=x, WR x (A[s] (=) A,                          SAME AS data struct)
#  - A[t]|=x, WR x (A[s] (=) A[t]                          if s intersects t)

# WR A[s]
#  - RD A          (A[s] (=) A,                          SAME AS data struct)
#  - WR A          (detect 2-writer conflict,            SAME AS data struct)
#  - WR A[t]       (detect 2-writer conflict               if s intersects t)
#  - A   |=x       (mark A as writer in net {A,x},       SAME AS data struct)
#  - A   |=x, WR x (detect 2-writer conflict,            SAME AS data struct)
#  - A[t]|=x       (mark A[t] as writer in net {A[t],x}    if s intersects t)
#  - A[t]|=x, WR x (detect 2-writer conflict               if s intersects t)

# A[s]|=x
#  - WR A          (mark A[s] as writer in net {A[s],x}, SAME AS data struct)
#  - A|=y, WR y    (mark A[s] as writer in net {A[s],x}, SAME AS data struct)
#  - A[t]|=y, WR y (mark A[s] as writer in net {A[s],x},   if s intersects t)

# A[s]|=x, WR x
#  - RD A          (A[s] (=) A,                          SAME AS data struct)
#  - WR A          (detect 2-writer conflict,            SAME AS data struct)
#  - A   |=y       (mark A as writer in net {A,y}        SAME AS data struct)
#  - A   |=y, WR y (detect 2-writer conflict,            SAME AS data struct)
#  - A[t]|=y, WR y (detect 2-writer conflict               if s intersects t)

# RD A[s] - WR A
def test_rd_As_wr_A_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_A():
        s.A @= Bits32( 123 )

      @update
      def up_rd_As():
        assert s.A[0:16] == 123

  _test_model( Top )

# RD A[s] - WR A[t], intersect
def test_rd_As_wr_At_impl_intersect():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_At():
        s.A[8:24] @= Bits16( 0xff )

      @update
      def up_rd_As():
        assert s.A[0:16] == 0xff00

  _test_model( Top )

# RD A[s] - WR A[t], not intersect
def test_rd_As_wr_At_impl_disjoint():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_At():
        s.A[16:32] @= Bits16( 0xff )

      @update
      def up_rd_As():
        assert s.A[0:16] == 0

  m = Top()
  m.elaborate()
  simple_sim_pass( m, 0x123 )

  # assert len(m._all_constraints) == 0 # no constraint at all!

# WR A[s] - WR A
def test_wr_As_wr_A_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_As():
        s.A[1:3] @= Bits2( 2 )

      @update
      def up_wr_A():
        s.A @= Bits32( 123 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# WR A[s] - WR A[t], intersect
def test_wr_As_wr_At_intersect():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_As():
        s.A[1:3] @= Bits2( 2 )

      @update
      def up_wr_At():
        s.A[2:4] @= Bits2( 2 )

      @update
      def up_rd_A():
        z = s.A

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# WR A[s] - WR A[t], not intersect
def test_wr_As_wr_At_disjoint():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_As():
        s.A[1:3] @= Bits2( 2 )

      @update
      def up_wr_At():
        s.A[5:7] @= Bits2( 2 )

      @update
      def up_rd_A():
        z = s.A

  _test_model( Top )

# WR A[s] - RD A
def test_wr_As_rd_A_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_As():
        s.A[1:3] @= Bits2( 2 )

      @update
      def up_rd_A():
        z = s.A

  _test_model( Top )

# WR A[s] - RD A, RD A[t], intersect
def test_wr_As_rd_A_rd_At_can_schedule():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_As():
        s.A[1:3] @= Bits2( 2 )

      @update
      def up_rd_A():
        z = s.A

      @update
      def up_rd_As():
        assert s.A[2:4] == 1

  _test_model( Top )

# WR A[s] - RD A, RD A[t], not intersect
def test_wr_As_rd_A_rd_At_cannot_schedule():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_As():
        s.A[1:3] @= Bits2( 2 )

      @update
      def up_rd_A():
        z = s.A

      @update
      def up_rd_At():
        assert s.A[3:5] == 0

  m = Top()
  m.elaborate()
  simple_sim_pass( m, 0x123 )

  # assert len(m._all_constraints) == 1
  # x, y = list(m._all_constraints)[0]

  # assert  m._all_id_upblk[x].__name__ == "up_wr_As" and \
          # m._all_id_upblk[y].__name__ == "up_rd_A" # only one constraint

# WR A - RD A[s], RD A[t]
def test_wr_A_rd_slices_can_schedule():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_A():
        s.A @= 0x12345678

      @update
      def up_rd_As():
        assert s.A[0:16] == 0x5678

      @update
      def up_rd_At():
        assert s.A[8:24] == 0x3456

  _test_model( Top )

# WR A[s] - RD A, RD A[t], not intersect
def test_wr_As_rd_A_rd_At_bit_cannot_schedule():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @update
      def up_wr_As():
        s.A[0:16] @= Bits16( 0x1234 )

      @update
      def up_rd_A():
        z = s.A

      @update
      def up_rd_At():
        assert s.A[16] == 0

  m = Top()
  m.elaborate()
  simple_sim_pass( m, 0x123 )

  # assert len(m._all_constraints) == 1
  # x, y = list(m._all_constraints)[0]

  # assert  m._all_id_upblk[x].__name__ == "up_wr_As" and \
          # m._all_id_upblk[y].__name__ == "up_rd_A" # only one constraint

# RD A[s] - A|=x, WR x
def test_connect_rd_As_wr_x_conn_A_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits32 )
      s.A  = Wire( Bits32 )

      connect( s.A, s.x )

      @update
      def up_wr_x():
        s.x @= Bits32( 123 )

      @update
      def up_rd_As():
        assert s.A[0:16] == 123

  _test_model( Top )

# RD A[s] - A[t]|=x, WR x, intersect
def test_connect_rd_As_wr_x_conn_At_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.A[0:24], s.x )

      @update
      def up_wr_x():
        s.x @= Bits24( 0x123456 )

      @update
      def up_rd_As():
        assert s.A[0:16] == 0x3456

  _test_model( Top )

# RD A[s] - A[t]|=x, WR x, not intersect
def test_connect_rd_As_wr_x_conn_At_disjoint():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.A[0:24], s.x )

      @update
      def up_wr_x():
        s.x @= Bits24( 0x123456 )

      @update
      def up_rd_As():
        assert s.A[24:32] == 0

  m = Top()
  m.elaborate()
  simple_sim_pass( m, 0x123 )

  # assert len(m._all_constraints) == 1
  # x, y = list(m._all_constraints)[0]

  # assert  m._all_id_upblk[x].__name__ == "up_wr_x" and \
          # m._all_id_upblk[y].__name__ == "s_x__1" # connection block

# WR A[s] - A|=x
def test_connect_wr_As_rd_x_conn_A_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits32 )
      s.A  = Wire( Bits32 )

      connect( s.x, s.A )

      @update
      def up_wr_As():
        s.A[0:24] @= Bits24( 0x123456 )

  _test_model( Top )

# WR A[s] - A|=x, WR x
def test_connect_wr_As_wr_x_conn_A_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits32 )
      s.A  = Wire( Bits32 )

      connect( s.x, s.A )

      @update
      def up_wr_As():
        s.A[0:24] @= Bits24( 0x123456 )

      @update
      def up_wr_x():
        s.x @= Bits32( 0x87654321 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# WR A[s] - A[t]|=x, intersect
def test_connect_wr_As_rd_x_conn_At_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.x, s.A[8:32] )

      @update
      def up_wr_As():
        s.A[0:24] @= Bits24( 0x123456 )

  _test_model( Top )

# WR A[s] - A[t]|=x, not intersect
def test_connect_wr_As_rd_x_conn_At_no_driver():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.x, s.A[8:32] )

      @update
      def up_wr_As():
        s.A[0:4] @= Bits4( 0xf )

  try:
    _test_model( Top )
  except NoWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown NoWriterError.")

# WR A[s] - A[t]|=x, WR x, intersect
def test_connect_wr_As_wr_x_conn_At_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.x, s.A[8:32] )

      @update
      def up_wr_As():
        s.A[0:24] @= Bits24( 0x123456 )

      @update
      def up_wr_x():
        s.x @= Bits24( 0x654321 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# WR A[s] - A[t]|=x, WR x, not intersect
def test_connect_wr_As_wr_x_conn_At_disjoint():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.x, s.A[8:32] )

      @update
      def up_wr_As():
        s.A[0:4] @= Bits4( 0xf )

      @update
      def up_wr_x():
        s.x @= Bits24( 0x654321 )

      @update
      def up_rd_A():
        assert s.A == 0x6543210f

  _test_model( Top )

# A[s]|=x, WR x - RD A
def test_connect_wr_x_conn_As_rd_A_impl():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.A[8:32], s.x )

      @update
      def up_wr_x():
        s.x @= Bits24( 0x123456 )

      @update
      def up_rd_A():
        assert s.A == 0x12345600

  _test_model( Top )

# A[s]|=x, WR x - WR A
def test_connect_wr_x_conn_As_wr_A_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.y  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.A[8:32], s.x )
      connect( s.x, s.y )

      @update
      def up_wr_x():
        s.x @= Bits24( 0x123456 )

      @update
      def up_wr_A():
        s.A @= Bits32( 0x12345678 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# A[s]|=x - WR A
def test_connect_rd_x_conn_As_wr_A_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )

      connect( s.A[8:32], s.x )

      @update
      def up_wr_A():
        s.A @= Bits32( 0x12345678 )

      @update
      def up_rd_x():
        assert s.x == 0x123456

  _test_model( Top )

# A[s]|=x, WR x - A|=y, WR y
def test_connect_wr_x_conn_As_wr_y_conn_A_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )
      s.y  = Wire( Bits32 )

      connect( s.A[8:32], s.x )
      connect( s.A      , s.y )

      @update
      def up_wr_x():
        s.x @= Bits24( 0x123456 )

      @update
      def up_wr_y():
        s.y @= Bits32( 0x12345678 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# A[s]|=x, WR x - A[t]|=y, WR y, intersect
def test_connect_wr_x_conn_As_wr_y_conn_At_conflict():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )
      s.y  = Wire( Bits16 )

      connect( s.A[8:32], s.x )
      connect( s.A[0:16], s.y )

      @update
      def up_wr_x():
        s.x @= Bits24( 0x123456 )

      @update
      def up_wr_y():
        s.y @= Bits16( 0x1234 )

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

# A[s]|=x, WR x - A[t]|=y, WR y, not intersect
def test_connect_wr_x_conn_As_wr_y_conn_At_disjoint():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )
      s.y  = Wire( Bits4 )

      connect( s.A[8:32], s.x )
      connect( s.A[0:4],  s.y )

      @update
      def up_wr_x():
        s.x @= Bits24( 0x123456 )

      @update
      def up_wr_y():
        s.y @= Bits4( 0xf )

      @update
      def up_rd_A():
        assert s.A == 0x1234560f

  _test_model( Top )

# A[s]|=x, WR x - A|=y, RD y
def test_connect_wr_x_conn_As_rd_y_conn_A_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )
      s.y  = Wire( Bits32 )

      connect( s.A[8:32], s.x )
      connect( s.A,       s.y )

      @update
      def up_wr_x():
        s.x @= Bits24( 0x123456 )

      @update
      def up_rd_y():
        assert s.y == 0x12345600

  _test_model( Top )

# A[s]|=x - A|=y, WR y
def test_connect_rd_x_conn_As_wr_y_conn_A_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )
      s.y  = Wire( Bits32 )

      connect( s.A[8:32], s.x )
      connect( s.A,       s.y )

      @update
      def up_rd_x():
        assert s.x == 0x123456

      @update
      def up_wr_y():
        s.y @= Bits32( 0x12345678 )

  _test_model( Top )

# A[s]|=x - A[t]|=y, WR y, intersect
def test_connect_rd_x_conn_As_wr_y_conn_At_mark_writer():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x = Wire( Bits24 )
      s.A = Wire( Bits32 )
      s.y = Wire( Bits16 )

      connect( s.A[8:32], s.x )
      connect( s.A[0:16], s.y )

      @update
      def up_rd_x():
        assert s.x == 0x12

      @update
      def up_wr_y():
        s.y @= Bits16( 0x1234 )

  _test_model( Top )

# A[s]|=x - A[t]|=y, WR y, not intersect
def test_connect_rd_x_conn_As_wr_y_conn_no_driver():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.x  = Wire( Bits24 )
      s.A  = Wire( Bits32 )
      s.y  = Wire( Bits4 )

      connect( s.A[8:32], s.x )
      connect( s.A[0:4 ], s.y )

      @update
      def up_rd_x():
        assert s.x == 0

      @update
      def up_wr_y():
        s.y @= Bits4( 0xf )

  try:
    _test_model( Top )
  except NoWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown NoWriterError.")

def test_iterative_find_nets():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.w  = Wire( Bits32 )
      s.x  = Wire( Bits32 )
      s.y  = Wire( Bits32 )
      s.z  = Wire( Bits32 )

      connect( s.w[0:16],  s.x[8:24] ) # net1
      connect( s.x[16:32], s.y[0:16] ) # net2
      connect( s.y[8:24],  s.z[0:16] ) # net3

      @update
      def up_wr_s_w():
        s.w @= Bits32( 0x12345678 )

  _test_model( Top )

def test_multiple_sibling_slices():

  class Top( ComponentLevel3 ):
    def construct( s ):

      s.A  = Wire( Bits32 )
      s.x  = Wire( Bits16 )
      s.y  = Wire( Bits16 )
      s.z  = Wire( Bits16 )

      connect( s.A[0:16], s.x ) # net1

      connect( s.A[8:24], s.y ) # net2

      connect( s.A[16:32], s.z ) # net3

      @update
      def up_wr_s_w():
        s.x @= Bits16( 0x1234 )

  try:
    _test_model( Top )
  except NoWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown NoWriterError.")

def test_multiple_write_same_slice():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.out = Wire( Bits32 )
      s.wire = Wire( Bits32 )
      connect( s.out, s.wire[0:32] )

      @update
      def comb_upblk():
        s.wire[0:16]  @= 0
        s.wire[16:32] @= 1

  a = A()
  a.elaborate()

def test_multiple_write_same_slice_with_overlap():

  class A( ComponentLevel3 ):
    def construct( s ):
      s.out = Wire( Bits32 )
      s.wire = Wire( Bits32 )
      connect( s.out, s.wire[0:32] )

      s.wire2 = Wire( Bits24 )
      connect( s.wire2[0:24], s.wire[8:32] )

      @update
      def comb_upblk():
        s.wire[0:16]  @= 0
        s.wire2[0:24] @= 1

  try:
    _test_model( A )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")



#=========================================================================
# BehavioralRTLIRTypeCheckL4Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 30, 2019
"""Provide L4 behavioral RTLIR type check pass."""

from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt

from .BehavioralRTLIRTypeCheckL3Pass import (
    BehavioralRTLIRTypeCheckL3Pass,
    BehavioralRTLIRTypeCheckVisitorL3,
    BehavioralRTLIRTypeEnforcerL3,
)


class BehavioralRTLIRTypeCheckL4Pass( BehavioralRTLIRTypeCheckL3Pass ):
  def get_visitor_class( s ):
    return BehavioralRTLIRTypeCheckVisitorL4

class BehavioralRTLIRTypeCheckVisitorL4( BehavioralRTLIRTypeCheckVisitorL3 ):

  def __init__( s, component, freevars, accessed, tmpvars, rtlir_getter ):
    super().__init__( component, freevars, accessed, tmpvars, rtlir_getter )
    s.type_expect[ 'Attribute' ] = (
      ( 'value', (rt.Component, rt.Signal, rt.InterfaceView),
                 'the base of an attribute must be one of: module, struct, interface!' ),
    )

  def get_enforce_visitor( s ):
    return BehavioralRTLIRTypeEnforcerL4

  def visit_Attribute( s, node ):
    if isinstance( node.value.Type, rt.InterfaceView ):
      if not node.value.Type.has_property( node.attr ):
        raise PyMTLTypeError( s.blk, node.ast,
          f'{dtype.get_name()} does not have field {node.attr}!' )
      node.Type = node.value.Type.get_property( node.attr )
      # The attribute of an interface is always non-constant
      node._is_explicit = True
    else:
      super().visit_Attribute( node )

  def visit_Index( s, node ):
    if isinstance( node.value.Type, rt.Array ) and \
       isinstance( node.value.Type.get_sub_type(), rt.InterfaceView ):
      node.Type = node.value.Type.get_next_dim_type()
      node._is_explicit = True
      s._handle_index_extension( node, node.value, node.idx, 'index' )

    else:
      super().visit_Index( node )

#-------------------------------------------------------------------------
# Enforce types for all terms whose types are inferred (implicit)
#-------------------------------------------------------------------------

class BehavioralRTLIRTypeEnforcerL4( BehavioralRTLIRTypeEnforcerL3 ):
  pass



#=========================================================================
# TBGenPass_test.py
#=========================================================================
# Author : Shunning Jiang
# Date   : Mar 18, 2020
"""Test if the imported object works correctly."""

import pytest

from pymtl3 import DefaultPassGroup
from pymtl3.datatypes import Bits1, Bits32, Bits48, Bits64, clog2, mk_bits
from pymtl3.dsl import Component, InPort, Interface, OutPort, Placeholder, connect
from pymtl3.passes.backends.verilog import *
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.test_utils import TestVectorSimulator

from ...testcases import (
    CaseConnectArrayBits32Comp,
    CaseConnectArrayBits32FooIfcComp
)

def local_do_test( _m ):
  _m.elaborate()
  _m.apply( VerilogPlaceholderPass() )
  m = VerilogTranslationImportPass()( _m )
  m.set_metadata( VerilogTBGenPass.case_name, '1234' )
  m.apply( VerilogTBGenPass() )
  sim = TestVectorSimulator( m, _m._tv, _m._tv_in, _m._tv_out )
  sim.run_test()

def test_normal_queue( do_test ):
  # Test a Placeholder with params in `construct`
  def tv_in( m, tv ):
    m.enq_en  @= tv[0]
    m.enq_msg @= tv[1]
    m.deq_en  @= tv[3]
  def tv_out( m, tv ):
    if tv[2] != '*': assert m.enq_rdy == tv[2]
    if tv[4] != '*': assert m.deq_rdy == tv[5]
    if tv[5] != '*': assert m.deq_msg == tv[4]
  class VQueue( Component, Placeholder ):
    def construct( s, data_width, num_entries, count_width ):
      s.count   =  OutPort( mk_bits( count_width )  )
      s.deq_en  =  InPort( Bits1  )
      s.deq_rdy = OutPort( Bits1  )
      s.deq_msg = OutPort( mk_bits( data_width ) )
      s.enq_en  =  InPort( Bits1  )
      s.enq_rdy = OutPort( Bits1  )
      s.enq_msg =  InPort( mk_bits( data_width ) )
      s.set_metadata( VerilogTranslationImportPass.enable, True )
  num_entries = 1
  q = VQueue(
      data_width = 32,
      num_entries = num_entries,
      count_width = clog2(num_entries+1))
  tv = [
    #   enq                deq
    #   en    msg   rdy    en    msg   rdy
    [    1,    42,    1,    0,     0,    0  ],
    [    0,    43,    0,    1,    42,    1  ],
    [    1,    43,    1,    0,    42,    0  ],
    [    0,    44,    0,    1,    43,    1  ],
    [    1,    44,    1,    0,    43,    0  ],
    [    0,    45,    0,    1,    44,    1  ],
    [    1,    45,    1,    0,    44,    0  ],
  ]
  q._tv = tv
  q._tv_in = tv_in
  q._tv_out = tv_out
  do_test( q )

def test_CaseConnectArrayBits32FooIfcComp():
  case = CaseConnectArrayBits32FooIfcComp
  _m = case.DUT()
  _m.elaborate()
  _m.set_metadata( VerilogTranslationImportPass.enable, True )
  _m.apply( VerilogPlaceholderPass() )
  m = VerilogTranslationImportPass()( _m )
  m.set_metadata( VerilogTBGenPass.case_name, 'sb' )
  m.apply( VerilogTBGenPass() )
  sim = TestVectorSimulator( m, case.TV, case.TV_IN, case.TV_OUT )
  sim.run_test()

def test_CaseConnectArrayBits32Comp():
  case = CaseConnectArrayBits32Comp
  _m = case.DUT()
  _m.elaborate()
  _m.set_metadata( VerilogTranslationImportPass.enable, True )
  _m.apply( VerilogPlaceholderPass() )
  m = VerilogTranslationImportPass()( _m )
  m.set_metadata( VerilogTBGenPass.case_name, 'sb' )
  m.apply( VerilogTBGenPass() )
  sim = TestVectorSimulator( m, case.TV, case.TV_IN, case.TV_OUT )
  sim.run_test()



#=========================================================================
# errors.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 27, 2019
"""Exception classes for SystemVerilog backend."""

import inspect
import os
import sys
import traceback


class VerilogImportError( Exception ):
  """Error while performing import."""
  def __init__( self, obj, msg ):
    obj = str(obj)
    _, _, tb = sys.exc_info()

    if not tb:
      # There are no chaining exceptions
      return super().__init__(
        f"\nError trying to perform import on {obj}:\n- {msg}")
    else:
      # There is another cause for the current import error
      tb_info = traceback.extract_tb(tb)
      fname, line, func, _ = tb_info[-1]
      return super().__init__(
        f"\nIn file {fname}, Line {line}, Method {func}:"
        f"\nError trying to perform import on {obj}:\n- {msg}")

class VerilogTranslationError( Exception ):
  """SystemVerilog translation error."""
  def __init__( self, blk, _ast, msg ):
    ast = _ast.ast
    fname = os.path.abspath( inspect.getsourcefile( blk ) )
    line = inspect.getsourcelines( blk )[1]
    col = 0
    code = ""
    try:
      line += ast.lineno - 1
      col = ast.col_offset
      code_line = inspect.getsourcelines( blk )[0][ ast.lineno-1 ]
      code = '\n  ' + code_line.strip() + \
        '\n  '+ ' ' * (col-len(code_line)+len(code_line.lstrip())) + '^'
    except AttributeError:
      # The given AST node is neither expr nor stmt
      pass
    return super().__init__(
      f"\nIn file {fname}, Line {line}, Col {col}:{code}\n- {msg}" )

class VerilogStructuralTranslationError( Exception ):
  """SystemVerilog translation error."""
  def __init__( self, obj, msg ):
    return super().__init__(
      f"\nError while performing structural translation on {obj}\n- {msg}" )

class VerilogReservedKeywordError( Exception ):
  """SystemVerilog reserved keyword error."""
  def __init__( self, name, msg ):
    return super().__init__(
      f"- {name} is a SystemVerilog reserved keyword!\n- {msg}" )

class VerilogCloseLoopPurePythonSimError( Exception ):
  """SystemVerilog closed loop test error during pure python simulation."""
  def __init__( self, obj, msg ):
    return super().__init__(
      f"\nCalling closed-loop test method with component {obj}:\n- {msg}" )

class VerilogPlaceholderError( Exception ):
  """SystemVerilog placeholder error."""
  def __init__( self, obj, msg ):
    return super().__init__(
      f"\nError while pickling placeholder {obj}:\n- {msg}" )



"""
========================================================================
AddSimUtilFuncsPass.py
========================================================================

Author : Shunning Jiang
Date   : Jan 26, 2020
"""

import py

from pymtl3.datatypes import Bits, b1
from pymtl3.dsl.Component import Component
from pymtl3.dsl.Connectable import Const, Interface, MethodPort, Signal
from pymtl3.dsl.NamedObject import NamedObject
from pymtl3.extra.pypy import custom_exec
from pymtl3.passes.backends.verilog import VerilogTBGenPass
from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.errors import PassOrderError
from pymtl3.passes.tracing.CLLineTracePass import CLLineTracePass
from pymtl3.passes.tracing.LineTraceParamPass import LineTraceParamPass
from pymtl3.passes.tracing.PrintTextWavePass import PrintTextWavePass
from pymtl3.passes.tracing.VcdGenerationPass import VcdGenerationPass

from .SimpleTickPass import SimpleTickPass


class PrepareSimPass( BasePass ):
  def __init__( self, print_line_trace=True, reset_active_high=True ):
    assert reset_active_high in [ True, False ]

    self.print_line_trace  = print_line_trace
    self.reset_active_high = reset_active_high

  def __call__( self, top ):
    if hasattr(top, "sim_reset"):
      raise AttributeError( "Please rename the attribute top.sim_reset")
    if hasattr(top, "print_line_trace"):
      raise AttributeError( "Please modify the attribute top.print_line_trace")
    if not hasattr( top, "_sched" ):
      raise PassOrderError( "_sched" )
    if not hasattr( top._sched, "update_schedule" ):
      raise PassOrderError( "update_schedule" )
    if not hasattr( top._sched, "schedule_ff" ):
      raise PassOrderError( "schedule_ff" )
    if not hasattr( top._sched, "schedule_posedge_flip" ):
      raise PassOrderError( "schedule_posedge_flip" )

    top._sim = PassMetadata()

    self.create_print_line_trace( top )
    self.create_sim_cycle_count( top )
    self.create_lock_unlock_simulation( top )

    top.lock_in_simulation()

    self.create_sim_eval_comb( top )
    self.create_sim_tick( top )
    self.create_sim_reset( top )


  def create_sim_eval_comb( self, top ):
    # Pure RTL design, add eval_combinational
    if len( top.get_all_object_filter( lambda x: isinstance( x, MethodPort ) ) ) == 0 and \
       len( top.get_all_update_once() ) == 0:
      sim_eval_combinational = SimpleTickPass.gen_tick_function( [top._sim.check_top_level_inports] + top._sched.update_schedule )
    else:
      def sim_eval_combinational():
        raise NotImplementedError(f"top is not a pure RTL design. {'top'+repr(list(method_ports)[0])[1:]} is a method port.")

    top.sim_eval_combinational = sim_eval_combinational

  def create_sim_tick( self, top ):
    final_schedule = []

    # Pure RTL -- tick update blocks first
    if len( top.get_all_object_filter( lambda x: isinstance( x, MethodPort ) ) ) == 0 and \
       len( top.get_all_update_once() ) == 0:
      final_schedule = top._sched.update_schedule[::]

    if self.print_line_trace and hasattr( top, 'line_trace' ):
      final_schedule.append( top.print_line_trace )
    final_schedule += self.collect_ff_funcs( top )
    final_schedule += top._sched.update_schedule
    final_schedule.append( top._sim.check_top_level_inports )
    top.sim_tick = SimpleTickPass.gen_tick_function( final_schedule )

  def collect_ff_funcs( self, top ):
    # ff_funcs summarizes the execution at the clock edge
    ret = []
    # append tracing related work
    if top.has_metadata( VcdGenerationPass.vcd_func ):
      ret.append( top.get_metadata( VcdGenerationPass.vcd_func ) )

    if top.has_metadata( PrintTextWavePass.textwave_func ):
      ret.append( top.get_metadata( PrintTextWavePass.textwave_func ) )

    if top.has_metadata( VerilogTBGenPass.vtbgen_hooks ):
      ret.extend( top.get_metadata( VerilogTBGenPass.vtbgen_hooks ) )

    ret.extend( top._sched.schedule_ff )
    ret.extend( top._sched.schedule_posedge_flip )
    ret.append( self.create_advance_sim_cycle( top ) )

    # clear cl method flag after flip
    if top.has_metadata( CLLineTracePass.clear_cl_trace_func ):
      ret.append( top.get_metadata( CLLineTracePass.clear_cl_trace_func ) )

    return ret

  # Simulation related APIs
  def create_sim_reset( self, top ):
    ff = SimpleTickPass.gen_tick_function( self.collect_ff_funcs( top ) )
    up = SimpleTickPass.gen_tick_function( top._sched.update_schedule )

    print_line_trace = self.print_line_trace and hasattr( top, 'line_trace' )
    active_high      = self.reset_active_high

    def sim_reset():
      if print_line_trace:
        print()
      # cycle 0
      top.reset @= b1( active_high )
      up()

      ff()
      # cycle 1
      up()
      if print_line_trace:
        print( f"{top._sim.simulated_cycles:3}r {top.line_trace()}" )

      ff()
      # cycle 2
      up()
      if print_line_trace:
        print( f"{top._sim.simulated_cycles:3}r {top.line_trace()}" )

      ff()
      # cycle 3
      top.reset @= b1( not active_high )
      up()

    top.sim_reset = sim_reset

  def create_print_line_trace( self, top ):
    if self.print_line_trace and hasattr( top, 'line_trace' ):
      def print_line_trace():
        print( f"{top._sim.simulated_cycles:3}: {top.line_trace()}" )
      top.print_line_trace = print_line_trace

  @staticmethod
  def create_advance_sim_cycle( top ):
    top._sim.simulated_cycles = 0
    def advance_sim_cycle():
      top._sim.simulated_cycles += 1
    return advance_sim_cycle

  @staticmethod
  def create_sim_cycle_count( top ):
    def sim_cycle_count():
      return top._sim.simulated_cycles
    top.sim_cycle_count = sim_cycle_count

  @staticmethod
  def create_lock_unlock_simulation( top ):

    def lock_in_simulation():
      top._check_called_at_elaborate_top( "lock_in_simulation" )

      # Basically we want to avoid @= between elements in the same net since
      # we now use @=.
      # - First pass creates whole bunch of signals
      signal_object_mapping = {}

      Q = [ (top, top) ]
      while Q:
        current_obj, host = Q.pop()
        if isinstance( current_obj, list ):
          for i, obj in enumerate( current_obj ):
            if isinstance( obj, Signal ):
              try:
                value = obj.default_value()
                if obj._dsl.needs_double_buffer:
                  value <<= value
              except Exception as e:
                raise type(e)(str(e) + f' happens at {obj!r}')

              current_obj[i] = value

              signal_object_mapping[ obj ] = (current_obj, i, True, value)

            elif isinstance( obj, Component ):
              Q.append( (obj, obj) )
            elif isinstance( obj, (Interface, list) ):
              Q.append( (obj, host) )

        elif isinstance( current_obj, NamedObject ):
          for i, obj in current_obj.__dict__.items():
            if i[0] == '_': continue

            if isinstance( obj, Signal ):
              try:
                value = obj.default_value()
                if obj._dsl.needs_double_buffer:
                  value <<= value
              except Exception as e:
                raise type(e)(str(e) + f' happens at {obj!r}')

              setattr( current_obj, i, value )
              signal_object_mapping[obj] = (current_obj, i, False, value)

            elif isinstance( obj, Component ):
              Q.append( (obj, obj) )
            elif isinstance( obj, (Interface, list) ):
              Q.append( (obj, host) )

      # Swap all Signal objects with actual data
      nets = top.get_all_value_nets()

      # First step is to consolidate all non-slice signals in the same net
      # by pointing them to the same object
      # TODO optimize for bitstruct fields. Essentially only sliced signals
      # should be excluded.
      for writer, signals in nets:
        residence = None

        # Find the residence value
        if isinstance( writer, Const ) or writer.is_top_level_signal():
          residence = writer
        else:
          for x in signals:
            if x.is_top_level_signal():
              residence = x
              break

        if residence is None:
          continue # whole net is slice

        if isinstance( residence, Const ):
          residence_value = residence._dsl.const
        else:
          residence_value = signal_object_mapping[ residence ][-1]

        # Replace top-level signals in the net with residence value

        for x in signals:
          if x is not residence and x.is_top_level_signal():
            # swap old value with new residence value

            current_obj, i, is_list, value = signal_object_mapping[ x ]
            signal_object_mapping[ x ] = (current_obj, i, is_list, residence_value)

            if is_list:
              current_obj[i] = residence_value
            else:
              setattr( current_obj, i, residence_value )

      top._sim.signal_object_mapping = signal_object_mapping
      top._sim.locked_simulation = True

      # Add the function that checks if the Bits objects of
      # top-level input ports are modified. If so, it's mostly because
      # the top-level ports are assigned with = instead of @=.

      inports = []
      objs    = []
      for x in top._dsl.all_signals:
        if x.is_input_value_port() and x.is_top_level_signal() and x.get_host_component() is top:
          inports.append( x )
          objs.append( signal_object_mapping[x][-1] )

      src = """
def check_top_level_inports():
  {}
""".format( "\n  ".join([ f"assert {x} is obj{i}, 'Please use @= to assign top level InPort top.{repr(x)[2:]}'"
                            for i, x in enumerate(inports) ]) )
      _locals = {}
      _globals = { f"obj{i}" : x for i, x in enumerate(objs) }
      _globals['s'] = top
      custom_exec( py.code.Source(src).compile(), _globals, _locals)
      top._sim.check_top_level_inports = _locals['check_top_level_inports']

    def unlock_simulation():
      top._check_called_at_elaborate_top( "unlock_simulation" )
      try:
        assert top._sim.locked_simulation
      except:
        raise AttributeError("Cannot unlock an unlocked/never locked model.")

      top._sim.unlocked_simulation = True
      top._sim.locked_simulation = False

      # We will reuse the same Bits object since they won't change anymore
      for obj, (current_obj, i, is_list, _) in top._sim.signal_object_mapping.items():
        if is_list: current_obj[i] = obj
        else:       setattr( current_obj, i, obj )

    top.lock_in_simulation = lock_in_simulation
    top.unlock_simulation  = unlock_simulation



#=========================================================================
# test_utility.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 5, 2019
"""Provide utility methods for testing."""
import copy
from collections import deque

from hypothesis import strategies as st

from pymtl3.datatypes import Bits1, mk_bits
from pymtl3.dsl import OutPort
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.stdlib.test_utils import TestVectorSimulator

from ...yosys import YosysTranslationImportPass
from .. import VerilogTranslationImportPass

#=========================================================================
# test utility functions
#=========================================================================

def trim( s ):
  string = []
  lines = s.split( '\n' )
  for line in lines:
    _line = line.split()
    _string = "".join( _line )
    if _string and not _string.startswith( '//' ):
      string.append( "".join( line.split() ) )
  return "\n".join( string )

def check_eq( s, t ):
  if isinstance( s, list ) and isinstance( t, list ):
    for _s, _t in zip( s, t ):
      assert trim(_s) == trim(_t)
  else:
    assert trim(s) == trim(t)

#=========================================================================
# Hypothesis strategies for testing
#=========================================================================

def flatten( _rtype ):
  if isinstance( _rtype, rt.Array ):
    n_dim = _rtype.get_dim_sizes()
    rtype = _rtype.get_sub_type()
  else:
    n_dim = []
    rtype = _rtype
  return n_dim, rtype

#-------------------------------------------------------------------------
# Generate initialization data for a signal
#-------------------------------------------------------------------------

def VectorInitData( dtype ):
  nbits = dtype.get_length()
  min_val, max_val = -(2**(nbits-1)), 2**(nbits-1)-1
  value = 0
  return mk_bits( nbits )( value )

def StructInitData( dtype ):
  data = dtype.get_class()()
  for field_name, field in dtype.get_all_properties().items():
    setattr( data, field_name, DataTypeInitData( field ) )
  return data

def PackedArrayInitData( n_dim, sub_dtype ):
  if not n_dim:
    return DataTypeInitData( sub_dtype )
  else:
    data = []
    for i in range(n_dim[0]):
      data += [ PackedArrayInitData( n_dim[1:], sub_dtype ) ]
    return data

def DataTypeInitData( dtype ):
  if isinstance( dtype, rdt.Vector ):
    return VectorInitData( dtype )
  elif isinstance( dtype, rdt.Struct ):
    return StructInitData( dtype )
  elif isinstance( dtype, rdt.PackedArray ):
    n_dim = dtype.get_dim_sizes()
    sub_dtype = dtype.get_sub_dtype()
    return PackedArrayInitData( n_dim, sub_dtype )
  else:
    assert False, f"unrecognized data type {sub_dtype}!"

def InPortInitData( id_, port ):
  return { id_ : DataTypeInitData( port.get_dtype() ) }

def InterfaceInitData( id_, ifc ):
  init = {}
  for prop_name, prop_rtype in ifc.get_all_properties_packed():
    if isinstance( prop_rtype, rt.Array ):
      n_dim = prop_rtype.get_dim_sizes()
      sub_type = prop_rtype.get_sub_type()
      if isinstance(sub_type, rt.Port) and sub_type.get_direction() != "input":
        continue
      init.update( ArrayInitData( id_+"."+prop_name, n_dim, sub_type ) )
    elif isinstance( prop_rtype, rt.Port ):
      if prop_rtype.get_direction() == "input":
        init.update( InPortInitData( id_+"."+prop_name, prop_rtype ) )
    elif isinstance( prop_rtype, rt.InterfaceView ):
      init.update( InterfaceInitData( id_+"."+prop_name, prop_rtype ) )
  return init

def ArrayInitData( id_, n_dim, subtype ):
  if not n_dim:
    if isinstance( subtype, rt.Port ):
      return InPortInitData( id_, subtype )
    else:
      return InterfaceInitData( id_, subtype )
  else:
    init = {}
    for i in range(n_dim[0]):
      init.update( ArrayInitData(f'{id_}[{i}]', n_dim[1:], subtype) )
    return init

#-------------------------------------------------------------------------
# Hypothesis input data strategies
#-------------------------------------------------------------------------

@st.composite
def VectorDataStrategy( draw, dtype ):
  nbits = dtype.get_length()
  min_val, max_val = -(2**(nbits-1)), 2**(nbits-1)-1
  value = draw( st.integers( min_val, max_val ) )
  return mk_bits( nbits )( value )

@st.composite
def StructDataStrategy( draw, dtype ):
  data = dtype.get_class()()
  for field_name, field in dtype.get_all_properties().items():
    setattr( data, field_name, draw( DataTypeDataStrategy( field ) ) )
  return data

@st.composite
def PackedArrayDataStrategy( draw, n_dim, sub_dtype ):
  if not n_dim:
    return draw( DataTypeDataStrategy( sub_dtype ) )
  else:
    data = []
    for i in range(n_dim[0]):
      data += [ draw(PackedArrayDataStrategy( n_dim[1:], sub_dtype )) ]
    return data

@st.composite
def DataTypeDataStrategy( draw, dtype ):
  if isinstance( dtype, rdt.Vector ):
    return draw( VectorDataStrategy( dtype ) )
  elif isinstance( dtype, rdt.Struct ):
    return draw( StructDataStrategy( dtype ) )
  elif isinstance( dtype, rdt.PackedArray ):
    n_dim = dtype.get_dim_sizes()
    sub_dtype = dtype.get_sub_dtype()
    return draw( PackedArrayDataStrategy( n_dim, sub_dtype ) )
  else:
    assert False, f"unrecognized data type {sub_dtype}!"

@st.composite
def InPortDataStrategy( draw, id_, port ):
  return { id_ : draw(DataTypeDataStrategy( port.get_dtype() )) }

@st.composite
def InterfaceDataStrategy( draw, id_, ifc ):
  data = {}
  for prop_name, prop_rtype in ifc.get_all_properties_packed():
    if isinstance( prop_rtype, rt.Array ):
      n_dim = prop_rtype.get_dim_sizes()
      sub_type = prop_rtype.get_sub_type()
      if isinstance(sub_type, rt.Port) and sub_type.get_direction() != "input":
        continue
      data.update(draw(ArrayDataStrategy(id_+"."+prop_name, n_dim, sub_type)))
    elif isinstance( prop_rtype, rt.Port ):
      if prop_rtype.get_direction() == "input":
        data.update(draw(InPortDataStrategy(id_+"."+prop_name, prop_rtype)))
    elif isinstance( prop_rtype, rt.InterfaceView ):
      data.update(draw(InterfaceDataStrategy(id_+"."+prop_name, prop_rtype)))
  return data

@st.composite
def ArrayDataStrategy( draw, id_, n_dim, subtype ):
  if not n_dim:
    if isinstance( subtype, rt.Port ):
      return draw(InPortDataStrategy( id_, subtype ))
    else:
      return draw(InterfaceDataStrategy( id_, subtype ))
  else:
    data = {}
    for i in range(n_dim[0]):
      data.update(draw(
        ArrayDataStrategy(f'{id_}[{i}]', n_dim[1:], subtype)))
    return data

@st.composite
def DataStrategy( draw, dut ):
  """Return a strategy that generates input vector for component `dut`."""
  max_cycles = 10

  ret = []
  dut.elaborate()
  rifc = rt.RTLIRGetter(cache=False).get_component_ifc_rtlir( dut )
  ports = rifc.get_ports_packed()
  ifcs = rifc.get_ifc_views_packed()

  # Add reset cycle at the beginning
  reset1, reset2 = {}, {}
  for id_, port in ports:
    if id_ == "clk":
      reset1.update( { id_ : Bits1(0) } )
      reset2.update( { id_ : Bits1(1) } )
    elif id_ == "reset":
      reset1.update( { id_ : Bits1(1) } )
      reset2.update( { id_ : Bits1(1) } )
    else:
      n_dim, port_rtype = flatten( port )
      if port_rtype.get_direction() == "input":
        if n_dim:
          reset1.update( ArrayInitData( id_, n_dim, port_rtype ) )
          reset2.update( ArrayInitData( id_, n_dim, port_rtype ) )
        else:
          reset1.update( InPortInitData( id_, port_rtype ) )
          reset2.update( InPortInitData( id_, port_rtype ) )
  for id_, ifc in ifcs:
    n_dim, ifc_rtype = flatten( ifc )
    if n_dim:
      reset1.update( ArrayDataStrategy( id_, n_dim, ifc_rtype ) )
      reset2.update( ArrayDataStrategy( id_, n_dim, ifc_rtype ) )
    else:
      reset1.update( InterfaceInitData( id_, n_dim, ifc_rtype ) )
      reset2.update( InterfaceInitData( id_, n_dim, ifc_rtype ) )

  ret.append( reset1 )
  ret.append( reset2 )

  for i in range(max_cycles):
    data = {}
    for id_, port in ports:
      if id_ in [ "clk", "reset" ]:
        data.update( { id_ : Bits1(0) } )
      else:
        n_dim, port_rtype = flatten( port )
        if n_dim:
          if port_rtype.get_direction() == "input":
            data.update(draw( ArrayDataStrategy( id_, n_dim, port_rtype ) ))
        elif port_rtype.get_direction() == "input":
          data.update(draw( InPortDataStrategy( id_, port_rtype ) ))
    for id_, ifc in ifcs:
      n_dim, ifc_rtype = flatten( ifc )
      if n_dim:
        data.update(draw( ArrayDataStrategy( id_, n_dim, ifc_rtype ) ))
      else:
        data.update(draw( InterfaceDataStrategy( id_, ifc_rtype ) ))

    # Toggle clock signal
    toggle_data = {}
    for id_, signal in data.items():
      if id_ == "clk":
        toggle_data.update( { id_ : Bits1(1) } )
      else:
        toggle_data.update( { id_ : copy.deepcopy( signal ) } )

    ret.append( data )
    ret.append( toggle_data )

  return ret

#-------------------------------------------------------------------------
# closed_loop_component_input_test
#-------------------------------------------------------------------------

def closed_loop_component_input_test( dut, test_vector, tv_in, backend = "verilog" ):

  # Filter to collect all output ports of a component
  def outport_filter( obj ):
    return isinstance( obj, OutPort )

  assert backend in [ "verilog", "yosys" ], f"invalid backend {backend}!"

  dut.elaborate()
  reference_output = deque()
  all_output_ports = dut.get_local_object_filter( outport_filter )

  # Method to record reference outputs of the pure python component
  def ref_tv_out( model, test_vector ):
    dct = {}
    for out_port in all_output_ports:
      dct[ out_port ] = eval( "model." + out_port._dsl.my_name ).clone() # WE NEED TO CLONE NOW
    reference_output.append( dct )

  # Method to compare the outputs of the imported model and the pure python one
  def tv_out( model, test_vector ):
    assert len(reference_output) > 0, \
      "Reference runs for fewer cycles than the imported model!"
    for out_port in all_output_ports:
      ref = reference_output[0][out_port]
      imp = eval( "model." + out_port._dsl.my_name )
      assert ref == imp, f"Value mismatch: reference: {ref}, imported: {imp}"
    reference_output.popleft()

  # First simulate the pure python component to see if it has sane behavior
  reference_sim = TestVectorSimulator( dut, test_vector, tv_in, ref_tv_out )
  reference_sim.run_test()
  dut.unlock_simulation()

  # If it simulates correctly, translate it and import it back
  dut.elaborate()
  if backend == "verilog":
    dut.set_metadata( VerilogTranslationImportPass.enable, True )
    imported_obj = VerilogTranslationImportPass()( dut )
  elif backend == "yosys":
    dut.set_metadata( YosysTranslationImportPass.enable, True )
    imported_obj = YosysTranslationImportPass()( dut )

  # Run another vector simulator spin
  imported_sim = TestVectorSimulator( imported_obj, test_vector, tv_in, tv_out )
  imported_sim.run_test()

#-------------------------------------------------------------------------
# closed_loop_component_test
#-------------------------------------------------------------------------

def closed_loop_component_test( dut, data, backend = "verilog" ):
  """Test the DUT with the given test_vector.

  User who wish to use this method should pass in the hypothesis data
  strategy instance as `data`. This method will reflect on the interfaces
  and ports of the given DUT and generate input vector.
  """
  # Method to feed data into the DUT
  def tv_in( model, test_vector ):
    for name, data in test_vector.items():
      # `setattr` fails to set the correct value of an array if indexed by
      # a subscript. We use `exec` here to make sure the value of elements
      # are assigned correctly.
      exec( "model." + name + " @= data" )
  test_vector = data.draw( DataStrategy( dut ) )
  closed_loop_component_input_test( dut, test_vector, tv_in, backend )

#-------------------------------------------------------------------------
# closed_loop_test
#-------------------------------------------------------------------------

# TODO: A hypothesis test that works on generated test component AND
# generated input data.
def closed_loop_test():
  pass



from pymtl3 import *
from pymtl3.extra import clone_deepcopy
from .ifcs import RecvIfcRTL, SendIfcRTL

class RecvQueueAdapter( Component ):

  @non_blocking( lambda s: s.entry is not None )
  def deq( s ):
    ret = s.entry
    s.entry = None
    return ret

  def construct( s, Type ):
    s.recv = RecvIfcRTL( Type )
    s.entry = None

    @update_once
    def up_recv_rdy():
      s.recv.rdy @= (s.entry is None)

    @update_once
    def up_recv_msg():
      if (s.entry is None) & s.recv.val:
        s.entry = clone_deepcopy( s.recv.msg )

    s.add_constraints( M( s.deq )     < U( up_recv_rdy ), # deq before recv in a cycle -- pipe behavior
                       M( s.deq.rdy ) < U( up_recv_rdy ),
                       U( up_recv_rdy ) < U( up_recv_msg ) )


class SendQueueAdapter( Component ):

  @non_blocking( lambda s: s.entry is None )
  def enq( s, msg ):
    s.entry = clone_deepcopy( msg )

  def construct( s, Type ):
    s.send = SendIfcRTL( Type )

    s.entry = None
    s.sent  = Wire()

    @update_once
    def up_send():
      if s.entry is None:
        s.send.val @= 0
      else:
        s.send.val @= 1
        s.send.msg @= s.entry

    @update_ff
    def up_sent():
      s.sent <<= s.send.val & s.send.rdy

    @update_once
    def up_clear():
      if s.sent: # constraints reverse this
        s.entry = None

    s.add_constraints(
      U( up_clear )  < M( s.enq ),
      U( up_clear )  < M( s.enq.rdy ),
      M( s.enq )     < U( up_send ),
      M( s.enq.rdy ) < U( up_send )
    )
